import aerosandbox.numpy as np
from aerosandbox.geometry.polygon import Polygon
from aerosandbox.geometry.airfoil.airfoil_families import (
    get_NACA_coordinates,
    get_UIUC_coordinates,
    get_file_coordinates
)
from aerosandbox.library.aerodynamics import transonic
from aerosandbox.modeling.splines.hermite import linear_hermite_patch, cubic_hermite_patch
from scipy import interpolate
from typing import Callable, Union, Any, Dict, List
import json
from pathlib import Path
import os


class Airfoil(Polygon):
    """
    An airfoil. See constructor docstring for usage details.
    """

    def __init__(self,
                 name: str = "Untitled",
                 coordinates: Union[None, str, Path, np.ndarray] = None,
                 **deprecated_keyword_arguments
                 ):
        """
        Creates an Airfoil object.

        Args:

            name: Name of the airfoil [string]. Can also be used to auto-generate coordinates; see docstring for
            `coordinates` below.

            coordinates: A representation of the coordinates that define the airfoil. Can be one of several types of
            input; the following sequence of operations is used to interpret the meaning of the parameter:

                If `coordinates` is an Nx2 array of the [x, y] coordinates that define the airfoil, these are used
                as-is. Points are expected to be provided in standard airfoil order:

                    * Points should start on the upper surface at the trailing edge, continue forward over the upper
                    surface, wrap around the nose, continue aft over the lower surface, and then end at the trailing
                    edge on the lower surface.

                    * The trailing edge need not be closed, but many analyses implicitly assume that this gap is small.

                    * Take care to ensure that the point at the leading edge of the airfoil, usually (0, 0),
                    is not duplicated.

                If `coordinates` is provided as a string, it assumed to be the filepath to a *.dat file containing
                the coordinates; we attempt to load coordinates from this.

                If the coordinates are not specified and instead left as None, the constructor will attempt to
                auto-populate the coordinates based on the `name` parameter provided, in the following order of
                priority:

                    * If `name` is a 4-digit NACA airfoil (e.g. "naca2412"), coordinates will be created based on the
                    analytical equation.

                    * If `name` is the name of an airfoil in the UIUC airfoil database (e.g. "s1223", "e216",
                    "dae11"), coordinates will be loaded from that. Note that the string you provide must be exactly
                    the name of the associated *.dat file in the UIUC database.

        """
        ### Handle the airfoil name
        self.name = name

        ### Handle the coordinates
        self.coordinates = None
        if coordinates is None:  # If no coordinates are given
            try:  # See if it's a NACA airfoil
                self.coordinates = get_NACA_coordinates(name=self.name)
            except (ValueError, NotImplementedError):
                try:  # See if it's in the UIUC airfoil database
                    self.coordinates = get_UIUC_coordinates(name=self.name)
                except FileNotFoundError:
                    pass
                except UnicodeDecodeError:
                    import warnings
                    warnings.warn(
                        f"Airfoil {self.name} was found in the UIUC airfoil database, but could not be parsed.\n"
                        f"Check for any non-Unicode-compatible characters in the file, or specify the airfoil "
                        f"coordinates yourself.",
                    )
        else:

            try:  # If coordinates is a string, assume it's a filepath to a .dat file
                self.coordinates = get_file_coordinates(filepath=coordinates)
            except (OSError, FileNotFoundError, TypeError, UnicodeDecodeError):
                try:
                    shape = coordinates.shape
                    assert len(shape) == 2
                    assert shape[0] == 2 or shape[1] == 2
                    if not shape[1] == 2:
                        coordinates = np.transpose(shape)

                    self.coordinates = coordinates
                except AttributeError:
                    pass

        if self.coordinates is None:
            import warnings
            warnings.warn(
                f"Airfoil {self.name} had no coordinates assigned, and could not parse the `coordinates` input!",
                UserWarning,
                stacklevel=2,
            )

        ### Handle deprecated keyword arguments
        if len(deprecated_keyword_arguments) > 0:
            import warnings
            warnings.warn(
                "The `generate_polars`, `CL_function`, `CD_function`, and `CM_function` keyword arguments to the "
                "Airfoil constructor will be deprecated in an upcoming release. Their functionality is replaced"
                "by `Airfoil.get_aero_from_neuralfoil()`, which is faster and has better properties for optimization.",
                DeprecationWarning
            )

            generate_polars = deprecated_keyword_arguments.get("generate_polars", False)
            CL_function = deprecated_keyword_arguments.get("CL_function", None)
            CD_function = deprecated_keyword_arguments.get("CD_function", None)
            CM_function = deprecated_keyword_arguments.get("CM_function", None)

            ### Handle getting default polars
            if generate_polars:
                self.generate_polars()
            else:

                from aerosandbox.library.aerodynamics.viscous import Cf_flat_plate

                def print_default_warning():
                    warnings.warn("\n".join([
                        "Warning: Using a placeholder aerodynamics model for this Airfoil!",
                        "It's highly recommended that you either:",
                        "\ta) Specify polar functions in the Airfoil constructor, or",
                        "\tb) Call Airfoil.generate_polars() to auto-generate these polar functions with XFoil."
                    ]), stacklevel=3)

                def default_CL_function(alpha, Re, mach=0, deflection=0):
                    """
                    Lift coefficient.
                    """
                    print_default_warning()
                    Cl_inc = np.pi * np.sind(2 * alpha)
                    beta = (1 - mach) ** 2

                    Cl = Cl_inc * beta
                    return Cl

                def default_CD_function(alpha, Re, mach=0, deflection=0):
                    """
                    Drag coefficient.
                    """
                    print_default_warning()
                    Cf = Cf_flat_plate(Re_L=Re, method="hybrid-sharpe-convex")

                    ### Form factor model from Raymer, "Aircraft Design". Section 12.5, Eq. 12.30
                    t_over_c = 0.12
                    FF = 1 + 2 * t_over_c * 100 * t_over_c ** 4

                    Cd_inc = 2 * Cf * FF * (
                            1 + (np.sind(alpha) * 180 / np.pi / 5) ** 2
                    )
                    beta = (1 - mach) ** 2

                    Cd = Cd_inc * beta
                    return Cd

                def default_CM_function(alpha, Re, mach=0, deflection=0):
                    """
                    Pitching moment coefficient, as measured about quarter-chord.
                    """
                    print_default_warning()
                    return np.zeros_like(alpha)

                self.CL_function = default_CL_function
                self.CD_function = default_CD_function
                self.CM_function = default_CM_function

            ### Overwrite any default polars with those provided
            if CL_function is not None:
                self.CL_function = CL_function

            if CD_function is not None:
                self.CD_function = CD_function

            if CM_function is not None:
                self.CM_function = CM_function

    def __repr__(self) -> str:
        return f"Airfoil {self.name} ({self.n_points()} points)"

    def to_kulfan_airfoil(self,
                          n_weights_per_side: int = 8,
                          N1: float = 0.5,
                          N2: float = 1.0,
                          normalize_coordinates: bool = True,
                          use_leading_edge_modification: bool = True,
                          ) -> "KulfanAirfoil":

        from aerosandbox.geometry.airfoil.kulfan_airfoil import KulfanAirfoil
        from aerosandbox.geometry.airfoil.airfoil_families import get_kulfan_parameters

        parameters = get_kulfan_parameters(
            coordinates=self.coordinates,
            n_weights_per_side=n_weights_per_side,
            N1=N1,
            N2=N2,
            normalize_coordinates=normalize_coordinates,
            use_leading_edge_modification=use_leading_edge_modification,
        )

        return KulfanAirfoil(
            name=self.name,
            lower_weights=parameters["lower_weights"],
            upper_weights=parameters["upper_weights"],
            leading_edge_weight=parameters["leading_edge_weight"],
            TE_thickness=parameters["TE_thickness"],
            N1=N1,
            N2=N2,
        )

    def generate_polars(self,
                        alphas=np.linspace(-13, 13, 27),
                        Res=np.geomspace(1e3, 1e8, 12),
                        cache_filename: str = None,
                        xfoil_kwargs: Dict[str, Any] = None,
                        unstructured_interpolated_model_kwargs: Dict[str, Any] = None,
                        include_compressibility_effects: bool = True,
                        transonic_buffet_lift_knockdown: float = 0.3,
                        make_symmetric_polars: bool = False,
                        ) -> None:
        """
        Generates airfoil polar surrogate models (CL, CD, CM functions) from XFoil data and assigns them in-place to
        this Airfoil's polar functions.

        In other words, when this function is run, the following functions will be added (or overwritten) to the instance:
            * Airfoil.CL_function(alpha, Re, mach)
            * Airfoil.CD_function(alpha, Re, mach)
            * Airfoil.CM_function(alpha, Re, mach)

        Where alpha is in degrees.

        Warning: In-place operation! Modifies this Airfoil object by setting Airfoil.CL_function, etc. to the new
        polars.

        Args:

            alphas: The range of alphas to sample from XFoil at. Given in degrees.

            Res: The range of Reynolds numbers to sample from XFoil at. Dimensionless.

            cache_filename: A path-like filename (ideally a "*.json" file) that can be used to cache the XFoil
                results, making it much faster to regenerate the results.

                * If the file does not exist, XFoil will be run, and a cache file will be created.

                * If the file does exist, XFoil will not be run, and the cache file will be read instead.

            xfoil_kwargs: Keyword arguments to pass into the AeroSandbox XFoil module. See the aerosandbox.XFoil
                constructor for options.

            unstructured_interpolated_model_kwargs: Keyword arguments to pass into the UnstructuredInterpolatedModels
                that contain the polars themselves. See the aerosandbox.UnstructuredInterpolatedModel constructor for
                options.

            include_compressibility_effects: Includes compressibility effects in the polars, such as wave drag,
                mach tuck, CL effects across normal shocks. Note that accuracy here is dubious in the transonic regime
                and above - you should really specify your own CL/CD/CM models

        Returns: None (in-place), adds the following functions to the instance:

            * Airfoil.CL_function(alpha, Re, mach)
            * Airfoil.CD_function(alpha, Re, mach)
            * Airfoil.CM_function(alpha, Re, mach)

        """
        if self.coordinates is None:
            raise ValueError("Cannot generate polars for an airfoil that you don't have the coordinates of!")

        ### Set defaults
        if xfoil_kwargs is None:
            xfoil_kwargs = {}
        if unstructured_interpolated_model_kwargs is None:
            unstructured_interpolated_model_kwargs = {}

        xfoil_kwargs = {  # See asb.XFoil for the documentation on these.
            "verbose"      : False,
            "max_iter"     : 20,
            "xfoil_repanel": True,
            **xfoil_kwargs
        }

        unstructured_interpolated_model_kwargs = {  # These were tuned heuristically as defaults!
            "resampling_interpolator_kwargs": {
                "degree"   : 0,
                # "kernel": "linear",
                "kernel"   : "multiquadric",
                "epsilon"  : 3,
                "smoothing": 0.01,
                # "kernel": "cubic"
            },
            **unstructured_interpolated_model_kwargs
        }

        ### Retrieve XFoil Polar Data from the cache, if it exists.
        data = None
        if cache_filename is not None:
            try:
                with open(cache_filename, "r") as f:
                    data = {
                        k: np.array(v)
                        for k, v in json.load(f).items()
                    }
            except FileNotFoundError:
                pass

        ### Analyze airfoil with XFoil, if needed
        if data is None:
            ### If a cache filename is given, ensure that the directory exists.
            if cache_filename is not None:
                os.makedirs(os.path.dirname(cache_filename), exist_ok=True)

            from aerosandbox.aerodynamics.aero_2D import XFoil

            def get_run_data(Re):  # Get the data for an XFoil alpha sweep at one specific Re.
                run_data = XFoil(
                    airfoil=self,
                    Re=Re,
                    **xfoil_kwargs
                ).alpha(alphas)
                run_data["Re"] = Re * np.ones_like(run_data["alpha"])
                return run_data  # Data is a dict where keys are figures of merit [str] and values are 1D ndarrays.

            from tqdm import tqdm

            run_datas = [  # Get a list of dicts, where each dict is the result of an XFoil run at a particular Re.
                get_run_data(Re)
                for Re in tqdm(
                    Res,
                    desc=f"Running XFoil to generate polars for Airfoil '{self.name}':",
                )
            ]
            data = {  # Merge the dicts into one big database of all runs.
                k: np.concatenate(
                    tuple([run_data[k] for run_data in run_datas])
                )
                for k in run_datas[0].keys()
            }

            if make_symmetric_polars:  # If the airfoil is known to be symmetric, duplicate all data across alpha.
                keys_symmetric_across_alpha = ['CD', 'CDp', 'Re']  # Assumes the rest are antisymmetric

                data = {
                    k: np.concatenate([v, v if k in keys_symmetric_across_alpha else -v])
                    for k, v in data.items()
                }

            if cache_filename is not None:  # Cache the accumulated data for later use, if it doesn't already exist.
                with open(cache_filename, "w+") as f:
                    json.dump(
                        {k: v.tolist() for k, v in data.items()},
                        f,
                        indent=4
                    )

        ### Save the raw data as an instance attribute for later use
        self.xfoil_data = data

        ### Make the interpolators for attached aerodynamics
        from aerosandbox.modeling import UnstructuredInterpolatedModel

        attached_alphas_to_use = (
            alphas[::2] if len(alphas) > 20 else alphas
        )

        alpha_resample = np.concatenate([
            np.linspace(-180, attached_alphas_to_use.min(), 10)[:-1],
            attached_alphas_to_use,
            np.linspace(attached_alphas_to_use.max(), 180, 10)[1:],
        ])  # This is the list of points that we're going to resample from the XFoil runs for our InterpolatedModel, using an RBF.
        Re_resample = np.concatenate([
            Res.min() / 10 ** np.arange(1, 5)[::-1],
            Res,
            Res.max() * 10 ** np.arange(1, 5),
        ])  # This is the list of points that we're going to resample from the XFoil runs for our InterpolatedModel, using an RBF.

        x_data = {
            "alpha": data["alpha"],
            "ln_Re": np.log(data["Re"]),
        }
        x_data_resample = {
            "alpha": alpha_resample,
            "ln_Re": np.log(Re_resample)
        }

        CL_attached_interpolator = UnstructuredInterpolatedModel(
            x_data=x_data,
            y_data=data["CL"],
            x_data_resample=x_data_resample,
            **unstructured_interpolated_model_kwargs
        )
        log10_CD_attached_interpolator = UnstructuredInterpolatedModel(
            x_data=x_data,
            y_data=np.log10(data["CD"]),
            x_data_resample=x_data_resample,
            **unstructured_interpolated_model_kwargs
        )
        CM_attached_interpolator = UnstructuredInterpolatedModel(
            x_data=x_data,
            y_data=data["CM"],
            x_data_resample=x_data_resample,
            **unstructured_interpolated_model_kwargs
        )

        ### Determine if separated
        alpha_stall_positive = np.max(data["alpha"])  # Across all Re
        alpha_stall_negative = np.min(data["alpha"])  # Across all Re

        def separation_parameter(alpha, Re=0):
            """
            Positive if separated, negative if attached.

            This will be an input to a tanh() sigmoid blend via asb.numpy.blend(), so a value of 1 means the flow is
            ~90% separated, and a value of -1 means the flow is ~90% attached.
            """
            return 0.5 * np.softmax(
                alpha - alpha_stall_positive,
                alpha_stall_negative - alpha
            )

        ### Make the interpolators for separated aerodynamics
        from aerosandbox.aerodynamics.aero_2D.airfoil_polar_functions import airfoil_coefficients_post_stall

        CL_if_separated, CD_if_separated, CM_if_separated = airfoil_coefficients_post_stall(
            airfoil=self,
            alpha=alpha_resample
        )

        CD_if_separated = CD_if_separated + np.median(data["CD"])
        # The line above effectively ensures that separated CD will never be less than attached CD. Not exactly, but generally close. A good heuristic.

        CL_separated_interpolator = UnstructuredInterpolatedModel(
            x_data=alpha_resample,
            y_data=CL_if_separated
        )
        log10_CD_separated_interpolator = UnstructuredInterpolatedModel(
            x_data=alpha_resample,
            y_data=np.log10(CD_if_separated)
        )
        CM_separated_interpolator = UnstructuredInterpolatedModel(
            x_data=alpha_resample,
            y_data=CM_if_separated
        )

        def CL_function(alpha, Re, mach=0):

            alpha = np.mod(alpha + 180, 360) - 180  # Keep alpha in the valid range.
            CL_attached = CL_attached_interpolator({
                "alpha": alpha,
                "ln_Re": np.log(Re),
            })
            CL_separated = CL_separated_interpolator(alpha)  # Lift coefficient if separated

            CL_mach_0 = np.blend(  # Lift coefficient at mach = 0
                separation_parameter(alpha, Re),
                CL_separated,
                CL_attached
            )

            if include_compressibility_effects:
                prandtl_glauert_beta_squared_ideal = 1 - mach ** 2

                prandtl_glauert_beta = np.softmax(
                    prandtl_glauert_beta_squared_ideal,
                    -prandtl_glauert_beta_squared_ideal,
                    hardness=2.0  # Empirically tuned to data
                ) ** 0.5

                CL = CL_mach_0 / prandtl_glauert_beta

                mach_crit = transonic.mach_crit_Korn(
                    CL=CL,
                    t_over_c=self.max_thickness(),
                    sweep=0,
                    kappa_A=0.95
                )

                ### Accounts approximately for the lift drop due to buffet.
                buffet_factor = np.blend(
                    40 * (mach - mach_crit - (0.1 / 80) ** (1 / 3) - 0.06) * (mach - 1.1),
                    1,
                    transonic_buffet_lift_knockdown
                )

                ### Accounts for the fact that theoretical CL_alpha goes from 2 * pi (subsonic) to 4 (supersonic),
                # following linearized supersonic flow on a thin airfoil.
                cla_supersonic_ratio_factor = np.blend(
                    10 * (mach - 1),
                    4 / (2 * np.pi),
                    1,
                )

                return CL * buffet_factor * cla_supersonic_ratio_factor

            else:
                return CL_mach_0

        def CD_function(alpha, Re, mach=0):

            alpha = np.mod(alpha + 180, 360) - 180  # Keep alpha in the valid range.
            log10_CD_attached = log10_CD_attached_interpolator({
                "alpha": alpha,
                "ln_Re": np.log(Re),
            })
            log10_CD_separated = log10_CD_separated_interpolator(alpha)

            log10_CD_mach_0 = np.blend(
                separation_parameter(alpha, Re),
                log10_CD_separated,
                log10_CD_attached,
            )

            if include_compressibility_effects:

                CL_attached = CL_attached_interpolator({
                    "alpha": alpha,
                    "ln_Re": np.log(Re),
                })
                CL_separated = CL_separated_interpolator(alpha)

                CL_mach_0 = np.blend(
                    separation_parameter(alpha, Re),
                    CL_separated,
                    CL_attached
                )
                prandtl_glauert_beta_squared_ideal = 1 - mach ** 2

                prandtl_glauert_beta = np.softmax(
                    prandtl_glauert_beta_squared_ideal,
                    -prandtl_glauert_beta_squared_ideal,
                    hardness=2.0  # Empirically tuned to data
                ) ** 0.5

                CL = CL_mach_0 / prandtl_glauert_beta

                t_over_c = self.max_thickness()

                mach_crit = transonic.mach_crit_Korn(
                    CL=CL,
                    t_over_c=t_over_c,
                    sweep=0,
                    kappa_A=0.92
                )
                mach_dd = mach_crit + (0.1 / 80) ** (1 / 3)
                CD_wave = np.where(
                    mach < mach_crit,
                    0,
                    np.where(
                        mach < mach_dd,
                        20 * (mach - mach_crit) ** 4,
                        np.where(
                            mach < 0.97,
                            cubic_hermite_patch(
                                mach,
                                x_a=mach_dd,
                                x_b=0.97,
                                f_a=20 * (0.1 / 80) ** (4 / 3),
                                f_b=0.8 * t_over_c,
                                dfdx_a=0.1,
                                dfdx_b=0.8 * t_over_c * 8
                            ),
                            np.where(
                                mach < 1.1,
                                cubic_hermite_patch(
                                    mach,
                                    x_a=0.97,
                                    x_b=1.1,
                                    f_a=0.8 * t_over_c,
                                    f_b=0.8 * t_over_c,
                                    dfdx_a=0.8 * t_over_c * 8,
                                    dfdx_b=-0.8 * t_over_c * 8,
                                ),
                                np.blend(
                                    8 * 2 * (mach - 1.1) / (1.2 - 0.8),
                                    0.8 * 0.8 * t_over_c,
                                    1.2 * 0.8 * t_over_c,
                                )
                            )
                        )
                    )
                )

                # CD_wave = transonic.approximate_CD_wave(
                #     mach=mach,
                #     mach_crit=mach_crit,
                #     CD_wave_at_fully_supersonic=0.90 * self.max_thickness()
                # )

                return 10 ** log10_CD_mach_0 + CD_wave


            else:
                return 10 ** log10_CD_mach_0

        def CM_function(alpha, Re, mach=0):

            alpha = np.mod(alpha + 180, 360) - 180  # Keep alpha in the valid range.
            CM_attached = CM_attached_interpolator({
                "alpha": alpha,
                "ln_Re": np.log(Re),
            })
            CM_separated = CM_separated_interpolator(alpha)

            CM_mach_0 = np.blend(
                separation_parameter(alpha, Re),
                CM_separated,
                CM_attached
            )
            if include_compressibility_effects:
                prandtl_glauert_beta_squared_ideal = 1 - mach ** 2

                prandtl_glauert_beta = np.softmax(
                    prandtl_glauert_beta_squared_ideal,
                    -prandtl_glauert_beta_squared_ideal,
                    hardness=2.0  # Empirically tuned to data
                ) ** 0.5

                CM = CM_mach_0 / prandtl_glauert_beta

                return CM
            else:
                return CM_mach_0

        self.CL_function = CL_function
        self.CD_function = CD_function
        self.CM_function = CM_function

    def get_aero_from_neuralfoil(self,
                                 alpha: Union[float, np.ndarray],
                                 Re: Union[float, np.ndarray],
                                 mach: Union[float, np.ndarray] = 0.,
                                 n_crit: Union[float, np.ndarray] = 9.0,
                                 xtr_upper: Union[float, np.ndarray] = 1.0,
                                 xtr_lower: Union[float, np.ndarray] = 1.0,
                                 model_size: str = "large",
                                 control_surfaces: List["ControlSurface"] = None,
                                 include_360_deg_effects: bool = True,
                                 ) -> Dict[str, Union[float, np.ndarray]]:

        ### Normalize the inputs and evaluate
        normalization_outputs = self.normalize(return_dict=True)
        normalized_airfoil = normalization_outputs["airfoil"].to_kulfan_airfoil(
            n_weights_per_side=8,
            normalize_coordinates=False  # No need to redo this
        )
        delta_alpha = normalization_outputs["rotation_angle"]  # degrees
        x_translation_LE = normalization_outputs["x_translation"]
        y_translation_LE = normalization_outputs["y_translation"]
        scale = normalization_outputs["scale_factor"]

        x_translation_qc = -x_translation_LE + 0.25 * (1/scale * np.cosd(delta_alpha)) - 0.25
        y_translation_qc = -y_translation_LE + 0.25 * (1/scale * np.sind(-delta_alpha))

        raw_aero = normalized_airfoil.get_aero_from_neuralfoil(
            alpha=alpha + delta_alpha,
            Re=Re / scale,
            mach=mach,
            n_crit=n_crit,
            xtr_upper=xtr_upper,
            xtr_lower=xtr_lower,
            model_size=model_size,
            control_surfaces=control_surfaces,
            include_360_deg_effects=include_360_deg_effects
        )

        ### Correct the force vectors and lift-induced moment from translation
        extra_CM = -raw_aero["CL"] * x_translation_qc + raw_aero["CD"] * y_translation_qc
        raw_aero["CM"] = raw_aero["CM"] + extra_CM

        return raw_aero

    def plot_polars(self,
                    alphas: Union[np.ndarray, List[float]] = np.linspace(-20, 20, 500),
                    Res: Union[np.ndarray, List[float]] = 10 ** np.arange(3, 9),
                    mach: float = 0.,
                    show: bool = True,
                    Re_colors=None,
                    ) -> None:
        import matplotlib.pyplot as plt
        import aerosandbox.tools.pretty_plots as p

        fig, ax = plt.subplots(2, 2, figsize=(8, 7))
        plt.sca(ax[0, 0])
        plt.title("Lift Coefficient")
        plt.xlabel(r"Angle of Attack $\alpha$ [deg]")
        plt.ylabel(r"Lift Coefficient $C_L$")
        p.set_ticks(5, 1, 0.5, 0.1)
        plt.sca(ax[0, 1])
        plt.title("Drag Coefficient")
        plt.xlabel(r"Angle of Attack $\alpha$ [deg]")
        plt.ylabel(r"Drag Coefficient $C_D$")
        plt.ylim(bottom=0, top=0.05)
        p.set_ticks(5, 1, 0.01, 0.002)
        plt.sca(ax[1, 0])
        plt.title("Moment Coefficient")
        plt.xlabel(r"Angle of Attack $\alpha$ [deg]")
        plt.ylabel(r"Moment Coefficient $C_m$")
        p.set_ticks(5, 1, 0.05, 0.01)
        plt.sca(ax[1, 1])
        plt.title("Lift-to-Drag Ratio")
        plt.xlabel(r"Angle of Attack $\alpha$ [deg]")
        plt.ylabel(r"Lift-to-Drag Ratio $C_L/C_D$")
        p.set_ticks(5, 1, 20, 5)

        if Re_colors is None:
            Re_colors = p.mpl.colormaps.get_cmap('rainbow')(np.linspace(0, 1, len(Res)))
            Re_colors = [
                p.adjust_lightness(color, 0.7)
                for color in Re_colors
            ]

        for i, Re in enumerate(Res):
            kwargs = dict(
                alpha=alphas,
                Re=Re,
                mach=mach
            )

            plt.sca(ax[0, 0])
            plt.plot(
                alphas,
                self.CL_function(**kwargs),
                color=Re_colors[i],
                alpha=0.7
            )

            plt.sca(ax[0, 1])
            plt.plot(
                alphas,
                self.CD_function(**kwargs),
                color=Re_colors[i],
                alpha=0.7
            )

            plt.sca(ax[1, 0])
            plt.plot(
                alphas,
                self.CM_function(**kwargs),
                color=Re_colors[i],
                alpha=0.7
            )

            plt.sca(ax[1, 1])
            plt.plot(
                alphas,
                self.CL_function(**kwargs) / self.CD_function(**kwargs),
                color=Re_colors[i],
                alpha=0.7
            )

        from aerosandbox.tools.string_formatting import eng_string

        plt.sca(ax[0, 0])
        plt.legend(
            title="Reynolds Number",
            labels=[eng_string(Re) for Re in Res],
            ncol=2,
            # Note: `ncol` is old syntax; preserves backwards-compatibility with matplotlib 3.5.x.
            # New matplotlib versions use `ncols` instead.
            fontsize=8,
            loc='lower right'
        )

        if show:
            p.show_plot(
                f"Polar Functions for {self.name} Airfoil",
                legend=False,
            )

    def local_camber(self,
                     x_over_c: Union[float, np.ndarray] = np.linspace(0, 1, 101)
                     ) -> Union[float, np.ndarray]:
        """
        Returns the local camber of the airfoil at a given point or points.

        Args:
            x_over_c: The x/c locations to calculate the camber at [1D array, more generally, an iterable of floats]

        Returns:
            Local camber of the airfoil (y/c) [1D array].
        """
        upper = self.upper_coordinates()[::-1]
        lower = self.lower_coordinates()

        upper_interpolated = np.interp(
            x_over_c,
            upper[:, 0],
            upper[:, 1],
        )
        lower_interpolated = np.interp(
            x_over_c,
            lower[:, 0],
            lower[:, 1],
        )

        return (upper_interpolated + lower_interpolated) / 2

    def local_thickness(self,
                        x_over_c: Union[float, np.ndarray] = np.linspace(0, 1, 101)
                        ) -> Union[float, np.ndarray]:
        """
        Returns the local thickness of the airfoil at a given point or points.

        Args:
            x_over_c: The x/c locations to calculate the thickness at [1D array, more generally, an iterable of floats]

        Returns:
            Local thickness of the airfoil (y/c) [1D array].
        """
        upper = self.upper_coordinates()[::-1]
        lower = self.lower_coordinates()

        upper_interpolated = np.interp(
            x_over_c,
            upper[:, 0],
            upper[:, 1],
        )
        lower_interpolated = np.interp(
            x_over_c,
            lower[:, 0],
            lower[:, 1],
        )

        return upper_interpolated - lower_interpolated

    def max_camber(self,
                   x_over_c_sample: np.ndarray = np.linspace(0, 1, 101)
                   ) -> float:
        """
        Returns the maximum camber of the airfoil.

        Args:
            x_over_c_sample: Where should the airfoil be sampled to determine the max camber?

        Returns: The maximum thickness, as a fraction of chord.

        """
        return np.max(self.local_camber(x_over_c=x_over_c_sample))

    def max_thickness(self,
                      x_over_c_sample: np.ndarray = np.linspace(0, 1, 101)
                      ) -> float:
        """
        Returns the maximum thickness of the airfoil.

        Args:
            x_over_c_sample: Where should the airfoil be sampled to determine the max thickness?

        Returns: The maximum thickness, as a fraction of chord.

        """
        return np.max(self.local_thickness(x_over_c=x_over_c_sample))

    def draw(self,
             draw_mcl=False,
             draw_markers=True,
             backend="matplotlib",
             show=True
             ) -> None:
        """
        Draw the airfoil object.

        Args:
            draw_mcl: Should we draw the mean camber line (MCL)? [boolean]

            backend: Which backend should we use? "plotly" or "matplotlib"

            show: Should we show the plot? [boolean]

        Returns: None
        """
        x = np.reshape(np.array(self.x()), -1)
        y = np.reshape(np.array(self.y()), -1)
        if draw_mcl:
            x_mcl = np.linspace(np.min(x), np.max(x), len(x))
            y_mcl = self.local_camber(x_mcl)

        if backend == "matplotlib":
            import matplotlib.pyplot as plt
            import aerosandbox.tools.pretty_plots as p

            color = '#280887'
            plt.plot(
                x, y,
                ".-" if draw_markers else "-",
                zorder=11, color=color)
            plt.fill(x, y, zorder=10, color=color, alpha=0.2)
            if draw_mcl:
                plt.plot(x_mcl, y_mcl, "-", zorder=4, color=color, alpha=0.4)
            plt.axis("equal")
            if show:
                p.show_plot(
                    title=f"{self.name} Airfoil",
                    xlabel=r"$x/c$",
                    ylabel=r"$y/c$",

                )

        elif backend == "plotly":
            from aerosandbox.visualization.plotly import go
            fig = go.Figure()
            fig.add_trace(
                go.Scatter(
                    x=x,
                    y=y,
                    mode="lines+markers" if draw_markers else "lines",
                    name="Airfoil",
                    fill="toself",
                    line=dict(
                        color="blue"
                    )
                ),
            )
            if draw_mcl:
                fig.add_trace(
                    go.Scatter(
                        x=x_mcl,
                        y=y_mcl,
                        mode="lines",
                        name="Mean Camber Line (MCL)",
                        line=dict(
                            color="navy"
                        )
                    )
                )
            fig.update_layout(
                xaxis_title="x/c",
                yaxis_title="y/c",
                yaxis=dict(scaleanchor="x", scaleratio=1),
                title=f"{self.name} Airfoil"
            )
            if show:
                fig.show()
            else:
                return fig

    def LE_index(self) -> int:
        """
        Returns the index of the leading edge point in the airfoil coordinates.
        """
        return int(np.argmin(self.x()))

    def lower_coordinates(self) -> np.ndarray:
        """
        Returns an Nx2 ndarray of [x, y] coordinates that describe the lower surface of the airfoil.

        Order is from the leading edge to the trailing edge.

        Includes the leading edge point; be careful about duplicates if using this method in conjunction with
        Airfoil.upper_coordinates().
        """
        return self.coordinates[self.LE_index():, :]

    def upper_coordinates(self) -> np.ndarray:
        """
        Returns an Nx2 ndarray of [x, y] coordinates that describe the upper surface of the airfoil.

        Order is from the trailing edge to the leading edge.

        Includes the leading edge point; be careful about duplicates if using this method in conjunction with
        Airfoil.lower_coordinates().
        """
        return self.coordinates[:self.LE_index() + 1, :]

    def LE_radius(self, softness: float = 1e-6):
        LE_index = self.LE_index()

        # The three points closest to the leading edge
        LE_points = self.coordinates[LE_index - 1:LE_index + 2, :]

        # Make these 3 points into a triangle; these are the vectors representing edges
        edge_vectors = LE_points - np.roll(LE_points, 1, axis=0)
        edge_lengths = np.linalg.norm(edge_vectors, axis=1)

        # Now use a variant of Heron's formula for the circumcircle diameter
        a = edge_lengths[0]
        b = edge_lengths[1]
        c = edge_lengths[2]

        s = (a + b + c) / 2

        diameter = (a * b * c) / (
                2 * np.sqrt(s * (s - a) * (s - b) * (s - c))
        )

        return diameter / 2

    def TE_thickness(self) -> float:
        """
        Returns the thickness of the trailing edge of the airfoil.
        """
        x_gap = self.coordinates[0, 0] - self.coordinates[-1, 0]
        y_gap = self.coordinates[0, 1] - self.coordinates[-1, 1]

        return (
                x_gap ** 2 +
                y_gap ** 2
        ) ** 0.5

    def TE_angle(self) -> float:
        """
        Returns the trailing edge angle of the airfoil, in degrees.
        """
        upper_TE_vec = self.coordinates[0, :] - self.coordinates[1, :]
        lower_TE_vec = self.coordinates[-1, :] - self.coordinates[-2, :]

        return np.arctan2d(
            upper_TE_vec[0] * lower_TE_vec[1] - upper_TE_vec[1] * lower_TE_vec[0],
            upper_TE_vec[0] * lower_TE_vec[0] + upper_TE_vec[1] * upper_TE_vec[1]
        )

    # def LE_radius(self) -> float:
    #     """
    #     Gives the approximate leading edge radius of the airfoil, in chord-normalized units.
    #     """ # TODO finish me

    def repanel(self,
                n_points_per_side: int = 100,
                spacing_function_per_side=np.cosspace,
                ) -> 'Airfoil':
        """
        Returns a repaneled copy of the airfoil with cosine-spaced coordinates on the upper and lower surfaces.

        Args:

            n_points_per_side: Number of points per side (upper and lower) of the airfoil [int]

                Notes: The number of points defining the final airfoil will be `n_points_per_side * 2 - 1`,
                since one point (the leading edge point) is shared by both the upper and lower surfaces.

            spacing_function_per_side: Determines how to space the points on each side of the airfoil. Can be
                `np.linspace` or `np.cosspace`, or any other function of the call signature `f(a, b, n)` that returns
                a spaced array of `n` points between `a` and `b`. [function]

        Returns: A copy of the airfoil with the new coordinates.
        """

        old_upper_coordinates = self.upper_coordinates()  # Note: includes leading edge point, be careful about duplicates
        old_lower_coordinates = self.lower_coordinates()  # Note: includes leading edge point, be careful about duplicates

        # Find the streamwise distances between coordinates, assuming linear interpolation
        upper_distances_between_points = np.linalg.norm(np.diff(old_upper_coordinates, axis=0), axis=1)
        lower_distances_between_points = np.linalg.norm(np.diff(old_lower_coordinates, axis=0), axis=1)
        upper_distances_from_TE = np.concatenate(([0], np.cumsum(upper_distances_between_points)))
        lower_distances_from_LE = np.concatenate(([0], np.cumsum(lower_distances_between_points)))

        try:
            new_upper_coordinates = interpolate.CubicSpline(
                x=upper_distances_from_TE,
                y=old_upper_coordinates,
                axis=0,
                bc_type=(
                    (2, (0, 0)),
                    (1, (0, -1)),
                )
            )(spacing_function_per_side(0, upper_distances_from_TE[-1], n_points_per_side))

            new_lower_coordinates = interpolate.CubicSpline(
                x=lower_distances_from_LE,
                y=old_lower_coordinates,
                axis=0,
                bc_type=(
                    (1, (0, -1)),
                    (2, (0, 0)),
                )
            )(spacing_function_per_side(0, lower_distances_from_LE[-1], n_points_per_side))

        except ValueError as e:
            if not (
                    (np.all(np.diff(upper_distances_from_TE)) > 0) and
                    (np.all(np.diff(lower_distances_from_LE)) > 0)
            ):
                raise ValueError(
                    "It looks like your Airfoil has a duplicate point. Try removing the duplicate point and "
                    "re-running Airfoil.repanel()."
                )
            else:
                raise e

        return Airfoil(
            name=self.name,
            coordinates=np.concatenate((new_upper_coordinates, new_lower_coordinates[1:, :]), axis=0),
        )

    def normalize(
            self,
            return_dict: bool = False,
    ) -> Union['Airfoil', Dict[str, Union['Airfoil', float]]]:
        """
        Returns a copy of the Airfoil with a new set of `coordinates`, such that:
            - The leading edge (LE) is at (0, 0)
            - The trailing edge (TE) is at (1, 0)
            - The chord length is equal to 1

        The trailing-edge (TE) point is defined as the midpoint of the line segment connecting the first and last coordinate points (upper and lower surface TE points, respectively). The TE point is not necessarily one of the original points in the airfoil coordinates (`Airfoil.coordinates`); in general, it will not be one of the points if the TE thickness is nonzero.

        The leading-edge (LE) point is defined as the coordinate point with the largest Euclidian distance from the trailing edge. (In other words, if you were to center a circle on the trailing edge and progressively grow it, what's the last coordinate point that it would intersect?) The LE point is always one of the original points in the airfoil coordinates.

        The chord is defined as the Euclidian distance between the LE and TE points.

        Coordinate modifications to achieve the constraints described above (LE @ origin, TE at (1, 0), and chord of 1) are done by means of a translation and rotation.

        Args:

            return_dict: Determines the output type of the function.
                - If `False` (default), returns a copy of the Airfoil with the new coordinates.
                - If `True`, returns a dictionary with keys:
                
                        - "airfoil": a copy of the Airfoil with the new coordinates

                        - "x_translation": the amount by which the airfoil's LE was translated in the x-direction

                        - "y_translation": the amount by which the airfoil's LE was translated in the y-direction

                        - "scale_factor": the amount by which the airfoil was scaled (if >1, the airfoil had to get
                            bigger)

                        - "rotation_angle": the angle (in degrees) by which the airfoil was rotated about the LE.
                            Sign convention is that positive angles rotate the airfoil counter-clockwise.

                    All of thes values represent the "required change", e.g.:

                        - "x_translation" is the amount by which the airfoil's LE had to be translated in the
                            x-direction to get it to the origin.

                        - "rotation_angle" is the angle (in degrees) by which the airfoil had to be rotated (CCW).

        Returns: Depending on the value of `return_dict`, either:

            - A copy of the airfoil with the new coordinates (default), or

            - A dictionary with keys "airfoil", "x_translation", "y_translation", "scale_factor", and "rotation_angle".
                documentation for `return_tuple` for more information.
        """

        ### Step 1: Translate so that the LE point is at (0, 0).
        x_te = (self.x()[0] + self.x()[-1]) / 2
        y_te = (self.y()[0] + self.y()[-1]) / 2

        distance_to_te = (
                                 (self.x() - x_te) ** 2 +
                                 (self.y() - y_te) ** 2
                         ) ** 0.5

        le_index = np.argmax(distance_to_te)

        x_translation = -self.x()[le_index]
        y_translation = -self.y()[le_index]

        newfoil = self.translate(
            translate_x=x_translation,
            translate_y=y_translation,
        )

        ### Step 2: Scale so that the chord length is 1.
        scale_factor = 1 / distance_to_te[le_index]

        newfoil = newfoil.scale(
            scale_x=scale_factor,
            scale_y=scale_factor,
        )

        ### Step 3: Rotate so that the trailing edge is at (1, 0).

        x_te = (newfoil.x()[0] + newfoil.x()[-1]) / 2
        y_te = (newfoil.y()[0] + newfoil.y()[-1]) / 2

        rotation_angle = -np.arctan2(y_te, x_te)

        newfoil = newfoil.rotate(
            angle=rotation_angle,
        )

        if not return_dict:
            return newfoil
        else:
            return {
                "airfoil"       : newfoil,
                "x_translation" : x_translation,
                "y_translation" : y_translation,
                "scale_factor"  : scale_factor,
                "rotation_angle": np.degrees(rotation_angle),
            }

    def add_control_surface(
            self,
            deflection: float = 0.,
            hinge_point_x: float = 0.75,
            modify_coordinates: bool = True,
            modify_polars: bool = True,
    ) -> 'Airfoil':
        """
        Returns a version of the airfoil with a trailing-edge control surface added at a given point. Implicitly
        repanels the airfoil as part of this operation.

        Args:
            deflection: Deflection angle [degrees]. Downwards-positive.
            hinge_point_x: Chordwise location of the hinge, as a fraction of chord (x/c) [float]

        Returns: an Airfoil object with the new control deflection.

        """
        if modify_coordinates:
            # Find the hinge point
            hinge_point_y = np.where(
                deflection > 0,
                self.local_camber(hinge_point_x) - self.local_thickness(hinge_point_x) / 2,
                self.local_camber(hinge_point_x) + self.local_thickness(hinge_point_x) / 2,
            )

            # hinge_point_y = self.local_camber(hinge_point_x)
            hinge_point = np.reshape(
                np.array([hinge_point_x, hinge_point_y]),
                (1, 2)
            )

            def is_behind_hinge(xy: np.ndarray) -> np.ndarray:
                return (
                        (xy[:, 0] - hinge_point_x) * np.cosd(deflection / 2) -
                        (xy[:, 1] - hinge_point_y) * np.sind(deflection / 2)
                        > 0
                )

            orig_u = self.upper_coordinates()
            orig_l = self.lower_coordinates()[1:, :]

            rotation_matrix = np.rotation_matrix_2D(
                angle=-np.radians(deflection),
            )

            def T(xy):
                return np.transpose(xy)

            hinge_point_u = np.tile(hinge_point, (np.length(orig_u), 1))
            hinge_point_l = np.tile(hinge_point, (np.length(orig_l), 1))

            rot_u = T(rotation_matrix @ T(orig_u - hinge_point_u)) + hinge_point_u
            rot_l = T(rotation_matrix @ T(orig_l - hinge_point_l)) + hinge_point_l

            coordinates_x = np.concatenate([
                np.where(
                    is_behind_hinge(rot_u),
                    rot_u[:, 0],
                    orig_u[:, 0]
                ),
                np.where(
                    is_behind_hinge(rot_l),
                    rot_l[:, 0],
                    orig_l[:, 0]
                )
            ])
            coordinates_y = np.concatenate([
                np.where(
                    is_behind_hinge(rot_u),
                    rot_u[:, 1],
                    orig_u[:, 1]
                ),
                np.where(
                    is_behind_hinge(rot_l),
                    rot_l[:, 1],
                    orig_l[:, 1]
                )
            ])

            coordinates = np.stack([
                coordinates_x,
                coordinates_y
            ], axis=1)
        else:
            coordinates = self.coordinates

        if modify_polars:
            effectiveness = 1 - np.maximum(0, hinge_point_x + 1e-16) ** 2.751428551177291
            dalpha = deflection * effectiveness

            def CL_function(alpha: float, Re: float, mach: float) -> float:
                return self.CL_function(
                    alpha=alpha + dalpha,
                    Re=Re,
                    mach=mach,
                )

            def CD_function(alpha: float, Re: float, mach: float) -> float:
                return self.CD_function(
                    alpha=alpha + dalpha,
                    Re=Re,
                    mach=mach,
                )

            def CM_function(alpha: float, Re: float, mach: float) -> float:
                return self.CM_function(
                    alpha=alpha + dalpha,
                    Re=Re,
                    mach=mach,
                )

        else:
            CL_function = self.CL_function
            CD_function = self.CD_function
            CM_function = self.CM_function

        return Airfoil(
            name=self.name,
            coordinates=coordinates,
            CL_function=CL_function,
            CD_function=CD_function,
            CM_function=CM_function,
        )

    def set_TE_thickness(self,
                         thickness: float = 0.,
                         ) -> 'Airfoil':
        """
        Creates a modified copy of the Airfoil that has a specified trailing-edge thickness.

        Note that the trailing-edge thickness is given nondimensionally (e.g., as a fraction of chord).

        Args:
            thickness: The target trailing-edge thickness, given nondimensionally (e.g., as a fraction of chord).

        Returns: The modified airfoil.

        """
        ### Compute existing trailing-edge properties
        x_gap = self.coordinates[0, 0] - self.coordinates[-1, 0]
        y_gap = self.coordinates[0, 1] - self.coordinates[-1, 1]

        s_gap = (
                        x_gap ** 2 +
                        y_gap ** 2
                ) ** 0.5

        s_adjustment = (thickness - self.TE_thickness()) / 2

        ### Determine how much the trailing edge should move by in X and Y.
        if s_gap != 0:
            x_adjustment = s_adjustment * x_gap / s_gap
            y_adjustment = s_adjustment * y_gap / s_gap
        else:
            x_adjustment = 0
            y_adjustment = s_adjustment

        ### Decompose the existing airfoil coordinates to upper and lower sides, and x and y.
        u = self.upper_coordinates()
        ux = u[:, 0]
        uy = u[:, 1]

        le_x = ux[-1]

        l = self.lower_coordinates()[1:]
        lx = l[:, 0]
        ly = l[:, 1]

        te_x = (ux[0] + lx[-1]) / 2

        ### Create modified versions of the upper and lower coordinates
        new_u = np.stack(
            arrays=[
                ux + x_adjustment * (ux - le_x) / (te_x - le_x),
                uy + y_adjustment * (ux - le_x) / (te_x - le_x)
            ],
            axis=1
        )
        new_l = np.stack(
            arrays=[
                lx - x_adjustment * (lx - le_x) / (te_x - le_x),
                ly - y_adjustment * (lx - le_x) / (te_x - le_x)
            ],
            axis=1
        )

        ### If the desired thickness is zero, ensure that is precisely reached.
        if thickness == 0:
            new_l[-1] = new_u[0]

        ### Combine the upper and lower surface coordinates into a single array.
        new_coordinates = np.concatenate(
            [
                new_u,
                new_l
            ],
            axis=0
        )

        ### Return a new Airfoil with the desired coordinates.
        return Airfoil(
            name=self.name,
            coordinates=new_coordinates
        )

    def scale(self,
              scale_x: float = 1.,
              scale_y: float = 1.,
              ) -> 'Airfoil':
        """
        Scales an Airfoil about the origin.

        Args:

            scale_x: Amount to scale in the x-direction.

            scale_y: Amount to scale in the y-direction. Scaling by a negative y-value will result in coordinates
                being re-ordered such that the order of the coordinates is still correct (i.e., starts from the
                upper-surface trailing edge, continues along the upper surface to the nose, then continues along the
                lower surface to the trailing edge).

        Returns: A copy of the Airfoil with appropriate scaling applied.
        """
        x = self.x() * scale_x
        y = self.y() * scale_y

        if scale_x < 0:
            TE_index = np.argmax(x)
            x = np.concatenate([
                x[TE_index::-1],
                x[-2:TE_index-1:-1]
            ])
            y = np.concatenate([
                y[TE_index::-1],
                y[-2:TE_index-1:-1]
            ])

        if scale_y < 0:
            x = x[::-1]
            y = y[::-1]

        coordinates = np.stack((x, y), axis=1)

        return Airfoil(
            name=self.name,
            coordinates=coordinates,
        )

    def translate(self,
                  translate_x: float = 0.,
                  translate_y: float = 0.,
                  ) -> 'Airfoil':
        """
        Translates an Airfoil by a given amount.
        Args:
            translate_x: Amount to translate in the x-direction
            translate_y: Amount to translate in the y-direction

        Returns: The translated Airfoil.

        """
        x = self.x() + translate_x
        y = self.y() + translate_y

        return Airfoil(
            name=self.name,
            coordinates=np.stack((x, y), axis=1)
        )

    def rotate(self,
               angle: float,
               x_center: float = 0.,
               y_center: float = 0.
               ) -> 'Airfoil':
        """
        Rotates the airfoil clockwise by the specified amount, in radians.

        Rotates about the point (x_center, y_center), which is (0, 0) by default.

        Args:
            angle: Angle to rotate, counterclockwise, in radians.

            x_center: The x-coordinate of the center of rotation.

            y_center: The y-coordinate of the center of rotation.

        Returns: The rotated Airfoil.

        """

        coordinates = np.copy(self.coordinates)

        ### Translate
        translation = np.array([x_center, y_center])
        coordinates -= translation

        ### Rotate
        rotation_matrix = np.rotation_matrix_2D(
            angle=angle,
        )
        coordinates = (rotation_matrix @ coordinates.T).T

        ### Translate
        coordinates += translation

        return Airfoil(
            name=self.name,
            coordinates=coordinates
        )

    def blend_with_another_airfoil(self,
                                   airfoil: "Airfoil",
                                   blend_fraction: float = 0.5,
                                   n_points_per_side: int = 100,
                                   ) -> "Airfoil":
        """
        Blends this airfoil with another airfoil. Merges both the coordinates and the aerodynamic functions.

        Args:

            airfoil: The other airfoil to blend with.

            blend_fraction: The fraction of the other airfoil to use when blending. Defaults to 0.5 (50%).

                * A blend fraction of 0 will return an identical airfoil to this one (self).

                * A blend fraction of 1 will return an identical airfoil to the other one (`airfoil` parameter).

            n_points_per_side: The number of points per side to use when blending the coordinates of the two airfoils.

        Returns: A new airfoil that is a blend of this airfoil and another one.

        """
        foil_a = self.repanel(n_points_per_side=n_points_per_side)
        foil_b = airfoil.repanel(n_points_per_side=n_points_per_side)
        a_fraction = 1 - blend_fraction
        b_fraction = blend_fraction

        name = f"{a_fraction * 100:.0f}% {self.name}, {b_fraction * 100:.0f}% {airfoil.name}"

        coordinates = (
                a_fraction * foil_a.coordinates +
                b_fraction * foil_b.coordinates
        )

        return Airfoil(
            name=name,
            coordinates=coordinates,
        )

    # def normalize(self):
    #     pass  # TODO finish me

    def write_dat(self,
                  filepath: Union[str, Path] = None,
                  include_name: bool = True,
                  ) -> str:
        """
        Writes a .dat file corresponding to this airfoil to a filepath.

        Args:
            filepath: filepath (including the filename and .dat extension) [string]
                If None, this function returns the .dat file as a string.

            include_name: Should the name be included in the .dat file? (In a standard *.dat file, it usually is.)

        Returns: None

        """
        contents = []

        if include_name:
            contents += [self.name]

        contents += ["%f %f" % tuple(coordinate) for coordinate in self.coordinates]

        string = "\n".join(contents)

        if filepath is not None:
            with open(filepath, "w+") as f:
                f.write(string)

        return string

    # def get_xfoil_data(self,
    #                    a_start=-6,  # type: float
    #                    a_end=12,  # type: float
    #                    a_step=0.5,  # type: float
    #                    a_init=0,  # type: float
    #                    Re_start=1e4,  # type: float
    #                    Re_end=1e7,  # type: float
    #                    n_Res=30,  # type: int
    #                    mach=0,  # type: float
    #                    max_iter=20,  # type: int
    #                    repanel=False,  # type: bool
    #                    parallel=True,  # type: bool
    #                    verbose=True,  # type: bool
    #                    ):
    #     """ # TODO finish docstring
    #     Calculates aerodynamic performance data for a particular airfoil with XFoil.
    #     Does a 2D grid sweep of the alpha-Reynolds space at a particular Mach number.
    #     Populates two new instance variables:
    #         * self.xfoil_data_1D: A dict of XFoil data at all calculated operating points (1D arrays, NaNs removed)
    #         * self.xfoil_data_2D: A dict of XFoil data at all calculated operating points (2D arrays, NaNs present)
    #     :param a_start: Lower bound of angle of attack [deg]
    #     :param a_end: Upper bound of angle of attack [deg]
    #     :param a_step: Angle of attack increment size [deg]
    #     :param a_init: Angle of attack to initialize runs at. Should solve easily (0 recommended) [deg]
    #     :param Re_start: Reynolds number to begin sweep at. [unitless]
    #     :param Re_end: Reynolds number to end sweep at. [unitless]
    #     :param n_Res: Number of Reynolds numbers to sweep. Points are log-spaced.
    #     :param mach: Mach number to sweep at.
    #     :param max_iter: Maximum number of XFoil iterations per op-point.
    #     :param repanel: Should we interally repanel the airfoil within XFoil before running? [boolean]
    #         Consider disabling this if you try to do optimization based on this data (for smoothness reasons).
    #         Otherwise, it's generally a good idea to leave this on.
    #     :param parallel: Should we run in parallel? Generally results in significant speedup, but might not run
    #         correctly on some machines. Disable this if it's a problem. [boolean]
    #     :param verbose: Should we do verbose output? [boolean]
    #     :return: self (in-place operation that creates self.xfoil_data_1D and self.xfoil_data_2D)
    #     """
    #     assert a_init > a_start
    #     assert a_init < a_end
    #     assert Re_start < Re_end
    #     assert n_Res >= 1
    #     assert mach >= 0
    #
    #     Res = np.logspace(np.log10(Re_start), np.log10(Re_end), n_Res)
    #
    #     def get_xfoil_data_at_Re(Re):
    #
    #         import aerosandbox.numpy as np  # needs to be imported here to support parallelization
    #
    #         run_data_upper = self.xfoil_aseq(
    #             a_start=a_init + a_step,
    #             a_end=a_end,
    #             a_step=a_step,
    #             Re=Re,
    #             repanel=repanel,
    #             max_iter=max_iter,
    #             M=mach,
    #             reset_bls=True,
    #         )
    #         run_data_lower = self.xfoil_aseq(
    #             a_start=a_init,
    #             a_end=a_start,
    #             a_step=-a_step,
    #             Re=Re,
    #             repanel=repanel,
    #             max_iter=max_iter,
    #             M=mach,
    #             reset_bls=True,
    #         )
    #         run_data = {
    #             k: np.hstack((
    #                 run_data_lower[k][::-1],
    #                 run_data_upper[k]
    #             )) for k in run_data_upper.keys()
    #         }
    #         return run_data
    #
    #     if verbose:
    #         print("Running XFoil sweeps on Airfoil %s..." % self.name)
    #         import time
    #         start_time = time.time()
    #
    #     if not parallel:
    #         runs_data = [get_xfoil_data_at_Re(Re) for Re in Res]
    #     else:
    #         import multiprocess as mp
    #         pool = mp.Pool(mp.cpu_count())
    #         runs_data = pool.map(get_xfoil_data_at_Re, Res)
    #         pool.close()
    #
    #     if verbose:
    #         run_time = time.time() - start_time
    #         print("XFoil Runtime: %.3f sec" % run_time)
    #
    #     xfoil_data_2D = {}
    #     for k in runs_data[0].keys():
    #         xfoil_data_2D[k] = np.vstack([
    #             d[k]
    #             for d in runs_data
    #         ])
    #     xfoil_data_2D["Re"] = np.tile(Res, (
    #         xfoil_data_2D["alpha"].shape[1],
    #         1
    #     )).T
    #     np.place(
    #         arr=xfoil_data_2D["Re"],
    #         mask=np.isnan(xfoil_data_2D["alpha"]),
    #         vals=np.nan
    #     )
    #     xfoil_data_2D["alpha_indices"] = np.arange(a_start, a_end + a_step / 2, a_step)
    #     xfoil_data_2D["Re_indices"] = Res
    #
    #     self.xfoil_data_2D = xfoil_data_2D
    #
    #     # 1-dimensionalize it and remove NaNs
    #     xfoil_data_1D = {
    #         k: remove_nans(xfoil_data_2D[k].reshape(-1))
    #         for k in xfoil_data_2D.keys()
    #     }
    #     self.xfoil_data_1D = xfoil_data_1D
    #
    #     return self
    #
    # def has_xfoil_data(self, raise_exception_if_absent=True):
    #     """
    #     Runs a quick check to see if this airfoil has XFoil data.
    #     :param raise_exception_if_absent: Boolean flag to raise an Exception if XFoil data is not found.
    #     :return: Boolean of whether or not XFoil data is present.
    #     """
    #     data_present = (
    #             hasattr(self, 'xfoil_data_1D') and
    #             hasattr(self, 'xfoil_data_2D')
    #     )
    #     if not data_present and raise_exception_if_absent:
    #         raise Exception(
    #             """This Airfoil %s does not yet have XFoil data,
    #             so you can't run the function you've called.
    #             To get XFoil data, first call:
    #                 Airfoil.get_xfoil_data()
    #             which will perform an in-place update that
    #             provides the data.""" % self.name
    #         )
    #     return data_present
    #
    # def plot_xfoil_data_contours(self):
    #     self.has_xfoil_data()  # Ensure data is present.
    #     from matplotlib import colors
    #
    #     d = self.xfoil_data_1D  # data
    #
    #     fig = plt.figure(figsize=(10, 8), dpi=200)
    #
    #     ax = fig.add_subplot(311)
    #     coords = self.coordinates
    #     plt.plot(coords[:, 0], coords[:, 1], '.-', color='#280887')
    #     plt.xlabel(r"$x/c$")
    #     plt.ylabel(r"$y/c$")
    #     plt.title(r"XFoil Data for %s Airfoil" % self.name)
    #     plt.axis("equal")
    #
    #     with plt.style.context("default"):
    #         ax = fig.add_subplot(323)
    #         x = d["Re"]
    #         y = d["alpha"]
    #         z = d["Cl"]
    #         levels = np.linspace(-0.5, 1.5, 21)
    #         norm = None
    #         CF = ax.tricontourf(x, y, z, levels=levels, norm=norm, cmap="plasma", extend="both")
    #         C = ax.tricontour(x, y, z, levels=levels, norm=norm, colors='k', extend="both", linewidths=0.5)
    #         cbar = plt.colorbar(CF, format='%.2f')
    #         cbar.set_label(r"$C_l$")
    #         plt.grid(False)
    #         plt.xlabel(r"$Re$")
    #         plt.ylabel(r"$\alpha$")
    #         plt.title(r"$C_l$ from $Re$, $\alpha$")
    #         ax.set_xscale('log')
    #
    #         ax = fig.add_subplot(324)
    #         x = d["Re"]
    #         y = d["alpha"]
    #         z = d["Cd"]
    #         levels = np.logspace(-2.5, -1, 21)
    #         norm = colors.PowerNorm(gamma=1 / 2, vmin=np.min(levels), vmax=np.max(levels))
    #         CF = ax.tricontourf(x, y, z, levels=levels, norm=norm, cmap="plasma", extend="both")
    #         C = ax.tricontour(x, y, z, levels=levels, norm=norm, colors='k', extend="both", linewidths=0.5)
    #         cbar = plt.colorbar(CF, format='%.3f')
    #         cbar.set_label(r"$C_d$")
    #         plt.grid(False)
    #         plt.xlabel(r"$Re$")
    #         plt.ylabel(r"$\alpha$")
    #         plt.title(r"$C_d$ from $Re$, $\alpha$")
    #         ax.set_xscale('log')
    #
    #         ax = fig.add_subplot(325)
    #         x = d["Re"]
    #         y = d["alpha"]
    #         z = d["Cl"] / d["Cd"]
    #         x = x[d["alpha"] >= 0]
    #         y = y[d["alpha"] >= 0]
    #         z = z[d["alpha"] >= 0]
    #         levels = np.logspace(1, np.log10(150), 21)
    #         norm = colors.PowerNorm(gamma=1 / 2, vmin=np.min(levels), vmax=np.max(levels))
    #         CF = ax.tricontourf(x, y, z, levels=levels, norm=norm, cmap="plasma", extend="both")
    #         C = ax.tricontour(x, y, z, levels=levels, norm=norm, colors='k', extend="both", linewidths=0.5)
    #         cbar = plt.colorbar(CF, format='%.1f')
    #         cbar.set_label(r"$L/D$")
    #         plt.grid(False)
    #         plt.xlabel(r"$Re$")
    #         plt.ylabel(r"$\alpha$")
    #         plt.title(r"$L/D$ from $Re$, $\alpha$")
    #         ax.set_xscale('log')
    #
    #         ax = fig.add_subplot(326)
    #         x = d["Re"]
    #         y = d["alpha"]
    #         z = d["Cm"]
    #         levels = np.linspace(-0.15, 0, 21)  # np.logspace(1, np.log10(150), 21)
    #         norm = None  # colors.PowerNorm(gamma=1 / 2, vmin=np.min(levels), vmax=np.max(levels))
    #         CF = ax.tricontourf(x, y, z, levels=levels, norm=norm, cmap="plasma", extend="both")
    #         C = ax.tricontour(x, y, z, levels=levels, norm=norm, colors='k', extend="both", linewidths=0.5)
    #         cbar = plt.colorbar(CF, format='%.2f')
    #         cbar.set_label(r"$C_m$")
    #         plt.grid(False)
    #         plt.xlabel(r"$Re$")
    #         plt.ylabel(r"$\alpha$")
    #         plt.title(r"$C_m$ from $Re$, $\alpha$")
    #         ax.set_xscale('log')
    #
    #     plt.tight_layout()
    #     plt.show()
    #
    #     return self
    #
    # def plot_xfoil_data_all_polars(self,
    #                                n_lines_max=20,
    #                                Cd_plot_max=0.04,
    #                                ):
    #     """
    #     Plots the existing XFoil data found by running self.get_xfoil_data().
    #     :param n_lines_max: Maximum number of Reynolds numbers to plot. Useful if you ran a sweep with tons of Reynolds numbers.
    #     :param Cd_plot_max: Upper limit of Cd to plot [float]
    #     :return: self (makes plot)
    #     """
    #
    #     self.has_xfoil_data()  # Ensure data is present.
    #
    #     n_lines_max = min(n_lines_max, len(self.xfoil_data_2D["Re_indices"]))
    #
    #     fig, ax = plt.subplots(1, 1, figsize=(7, 6), dpi=200)
    #     indices = np.array(
    #         np.round(np.linspace(0, len(self.xfoil_data_2D["Re_indices"]) - 1, n_lines_max)),
    #         dtype=int
    #     )
    #     indices_worth_plotting = [
    #         np.min(remove_nans(self.xfoil_data_2D["Cd"][index, :])) < Cd_plot_max
    #         for index in indices
    #     ]
    #     indices = indices[indices_worth_plotting]
    #
    #     colors = plt.cm.rainbow(np.linspace(0, 1, len(indices)))[::-1]
    #     for i, Re in enumerate(self.xfoil_data_2D["Re_indices"][indices]):
    #         Cds = remove_nans(self.xfoil_data_2D["Cd"][indices[i], :])
    #         Cls = remove_nans(self.xfoil_data_2D["Cl"][indices[i], :])
    #         Cd_min = np.min(Cds)
    #         if Cd_min < Cd_plot_max:
    #             plt.plot(
    #                 Cds * 1e4,
    #                 Cls,
    #                 label="Re = %s" % eng_string(Re),
    #                 color=colors[i],
    #             )
    #     plt.xlim(0, Cd_plot_max * 1e4)
    #     plt.ylim(0, 2)
    #     plt.xlabel(r"$C_d \cdot 10^4$")
    #     plt.ylabel(r"$C_l$")
    #     plt.title("XFoil Polars for %s Airfoil" % self.name)
    #     plt.tight_layout()
    #     plt.legend()
    #     plt.show()
    #
    #     return self
    #
    # def plot_xfoil_data_polar(self,
    #                           Res,  # type: list
    #                           Cd_plot_max=0.04,
    #                           repanel=False,
    #                           parallel=True,
    #                           max_iter=40,
    #                           verbose=True,
    #                           ):
    #     """
    #     Plots CL-CD polar for a single Reynolds number or a variety of Reynolds numbers.
    #     :param Res: Reynolds number to plot polars at. Either a single float or an iterable (list, 1D ndarray, etc.)
    #     :param Cd_plot_max: Upper limit of Cd to plot [float]
    #     :param cl_step: Cl increment for XFoil runs. Trades speed vs. plot resolution. [float]
    #     :param repanel: Should we repanel the airfoil within XFoil? [boolean]
    #     :param parallel: Should we run different Res in parallel? [boolean]
    #     :param max_iter: Maximum number of iterations for XFoil to run. [int]
    #     :param verbose: Should we print information as we run the sweeps? [boolean]
    #     :return: self (makes plot)
    #     """
    #
    #     try:  # If it's not an iterable, make it one.
    #         Res[0]
    #     except TypeError:
    #         Res = [Res]
    #
    #     fig, ax = plt.subplots(1, 1, figsize=(7, 6), dpi=200)
    #     colors = plt.cm.rainbow(np.linspace(0, 1, len(Res)))[::-1]
    #
    #     def get_xfoil_data_at_Re(Re):
    #
    #         xfoil_data = self.xfoil_aseq(
    #             a_start=0,
    #             a_end=15,
    #             a_step=0.25,
    #             Re=Re,
    #             M=0,
    #             reset_bls=True,
    #             repanel=repanel,
    #             max_iter=max_iter,
    #             verbose=False,
    #         )
    #         Cd = remove_nans(xfoil_data["Cd"])
    #         Cl = remove_nans(xfoil_data["Cl"])
    #         return {"Cl": Cl, "Cd": Cd}
    #
    #     if verbose:
    #         print("Running XFoil sweeps...")
    #         import time
    #         start_time = time.time()
    #
    #     if not parallel:
    #         runs_data = [get_xfoil_data_at_Re(Re) for Re in Res]
    #     else:
    #         import multiprocess as mp
    #         pool = mp.Pool(mp.cpu_count())
    #         runs_data = pool.map(get_xfoil_data_at_Re, Res)
    #         pool.close()
    #
    #     if verbose:
    #         run_time = time.time() - start_time
    #         print("XFoil Runtime: %.3f sec" % run_time)
    #
    #     for i, Re in enumerate(Res):
    #         plt.plot(
    #             runs_data[i]["Cd"] * 1e4,
    #             runs_data[i]["Cl"],
    #             label="Re = %s" % eng_string(Re),
    #             color=colors[i],
    #         )
    #     plt.xlim(0, Cd_plot_max * 1e4)
    #     plt.ylim(0, 2)
    #     plt.xlabel(r"$C_d \cdot 10^4$")
    #     plt.ylabel(r"$C_l$")
    #     plt.title("XFoil Polars for %s Airfoil" % self.name)
    #     plt.tight_layout()
    #     plt.legend()
    #     plt.show()
    #
    #     return self


if __name__ == '__main__':
    af = Airfoil("dae11")

    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    fig, ax = plt.subplots(4, 2, figsize=(6.4, 6.4), dpi=200)

    alpha = np.linspace(-90, 90, 500)
    sizes = ["xxsmall", "xsmall", "small", "medium", "large", "xlarge", "xxlarge", "xxxlarge"]
    colors = plt.cm.rainbow(np.linspace(0, 1, len(sizes)))[::-1]

    for i, ms in enumerate(sizes):
        aero = af.get_aero_from_neuralfoil(
            alpha=alpha,
            Re=1e6,
            mach=0.3,
            model_size=ms,
        )

        kwargs = dict(
            alpha=0.5,
            color=colors[i],
        )
        for a, key in zip(ax.T.flatten(), ["CL", "CD", "CM", "Cpmin", "mach_crit", "Top_Xtr", "Bot_Xtr", "Cpmin_0"]):
            a.plot(alpha, aero[key], **kwargs)
            if key == "CD":
                a.set_yscale('log')
            a.set_ylabel(key)

    p.show_plot()

    # af.draw()
    # af.generate_polars(
    #     alphas=np.linspace(-10, 15, 61),
    # )
    # af.plot_polars(
    #     Res=np.geomspace(1e4, 1e6, 6)
    # )



import matplotlib.colors as mc
import matplotlib as mpl
import matplotlib.pyplot as plt
import aerosandbox.numpy as np
import colorsys

palettes = {
    "categorical": [
        "#4285F4",  # From Google logo, blue
        "#EA4335",  # From Google logo, red
        "#34A853",  # From Google logo, green
        "#ECB22E",  # From Slack logo, gold
        "#9467BD",  # From Matplotlib "tab10", purple
        "#8C564B",  # From Matplotlib "tab10", brown
        "#E377C2",  # From Matplotlib "tab10", pink
        "#7F7F7F",  # From Matplotlib "tab10", gray
    ],
    "pales": [
        "#C87A7A",
        "#B1866E",
        "#9B8E6E",
        "#90926E",
        "#85956D",
        "#77986D",
        "#6E9973",
        "#6F9882",
        "#719690",
        "#73959C",
        "#7692B1",
        "#868CBC",
    ]
}


def get_discrete_colors_from_colormap(
        cmap: str = "rainbow",
        N: int = 8,
        lower_bound: float = 0,
        upper_bound: float = 1,
):
    """
    Returns uniformly-spaced discrete color samples from a (continuous) colormap.
    Args:
        cmap: Name of a colormap.
        N: Number of colors to retrieve from colormap.

    Returns: A Nx4 array of (R, G, B, A) colors.

    """
    cmap = mpl.colormaps.get_cmap(cmap)
    colors = cmap(np.linspace(lower_bound, upper_bound, N))
    return colors


def adjust_lightness(color, amount=1.0):
    """
    Converts a color to HLS space, then multiplies the lightness by `amount`, then converts back to RGB.

    Args:
        color: A color, in any format that matplotlib understands.
        amount: The amount to multiply the lightness by. Valid range is 0 to infinity.

    Returns: A color, as an RGB tuple.

    """
    try:
        c = mc.cnames[color]
    except Exception:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))
    return colorsys.hls_to_rgb(c[0], max(0, min(1, amount * c[1])), c[2])


def get_last_line_color():
    lines = plt.gca().lines
    try:
        line = lines[-1]
        return line._color
    except IndexError:
        return palettes["categorical"][0] # TODO make this just the first color in the current palette



import aerosandbox as asb
import aerosandbox.numpy as np
import pandas as pd
import pytest

af = asb.Airfoil(
    "naca0012",
    CL_function=lambda alpha, Re, mach: 2 * np.pi * np.radians(alpha),
    CD_function=lambda alpha, Re, mach: np.zeros_like(alpha),
    CM_function=lambda alpha, Re, mach: np.zeros_like(alpha),
)

solvers = [
    # asb.AVL,
    asb.VortexLatticeMethod,
    asb.AeroBuildup
]


def test_horizontal():
    airplane_horizontal = asb.Airplane(
        wings=[
            asb.Wing(
                symmetric=True,
                color="blue",
                xsecs=[
                    asb.WingXSec(
                        xyz_le=[0, 0, 0],
                        chord=1,
                        airfoil=af
                    ),
                    asb.WingXSec(
                        xyz_le=[0, 1, 0],
                        chord=1,
                        airfoil=af
                    )
                ]
            )
        ]
    )
    # airplane_horizontal.draw_three_view()

    op_point_horiz = asb.OperatingPoint(
        velocity=100,
        alpha=5
    )

    horiz = pd.DataFrame(
        {
            solver.__name__: solver(airplane_horizontal, op_point_horiz).run()
            for solver in solvers
        }
    ).dropna().loc[["L", "CL", "CD"], :]

    print("\nHorizontal\n" + "-" * 80)
    print(horiz)

    assert horiz["VortexLatticeMethod"]['L'] == pytest.approx(
        horiz["AeroBuildup"]['L'],
        rel=0.1
    )


def test_v_tail():
    airplane_v_tail = asb.Airplane(
        wings=[
            asb.Wing(
                symmetric=True,
                color="purple",
                xsecs=[
                    asb.WingXSec(
                        xyz_le=[0, 0, 0],
                        chord=1,
                        airfoil=af
                    ),
                    asb.WingXSec(
                        xyz_le=[0, 1, 1],
                        chord=1,
                        airfoil=af
                    )
                ]
            )
        ]
    )

    # airplane_v_tail.draw_three_view()

    op_point_v_tail = asb.OperatingPoint(
        velocity=100,
        alpha=5,
        beta=5
    )

    v_tail = pd.DataFrame(
        {
            solver.__name__: solver(airplane_v_tail, op_point_v_tail).run()
            for solver in solvers
        }
    ).dropna().loc[["L", "CL", "Y", "CY", "CD"], :]

    print("\nV-Tail\n" + "-" * 80)
    print(v_tail)

    # TODO add fix for force direction, and add test


def test_vertical():
    airplane_vertical = asb.Airplane(
        wings=[
            asb.Wing(
                symmetric=True,
                color="red",
                xsecs=[
                    asb.WingXSec(
                        xyz_le=[0, 1, 0],
                        chord=1,
                        airfoil=af
                    ),
                    asb.WingXSec(
                        xyz_le=[0, 1, 1],
                        chord=1,
                        airfoil=af
                    )
                ]
            )
        ]
    )

    # airplane_vertical.draw_three_view()

    op_point_vert = asb.OperatingPoint(
        velocity=100,
        beta=5
    )

    vert = pd.DataFrame(
        {
            solver.__name__: solver(airplane_vertical, op_point_vert).run()
            for solver in solvers
        }
    ).dropna().loc[["Y", "CY", "CD"], :]

    print("\nVertical\n" + "-" * 80)
    print(vert)

    assert vert["VortexLatticeMethod"]['Y'] == pytest.approx(
        vert["AeroBuildup"]['Y'],
        rel=0.1
    )


if __name__ == '__main__':
    test_horizontal()
    test_v_tail()
    test_vertical()

    pytest.main()



import aerosandbox as asb
import aerosandbox.numpy as np
from typing import Type
import pytest

rtol = 0.10


def make_fuselage(
        alpha_geometric=0.,
        beta_geometric=0.
) -> asb.Fuselage:
    x = [0, 0.2, 0.8, 1]

    return asb.Fuselage(
        xsecs=[
            asb.FuselageXSec(
                xyz_c=[
                    xi * np.cosd(alpha_geometric) * np.cosd(beta_geometric),
                    xi * np.sind(beta_geometric),
                    xi * np.sind(-alpha_geometric)
                ],
                radius=np.where(
                    xi < 0.2,
                    0.25 * xi / 0.2,
                    np.where(
                        xi < 0.8,
                        0.25,
                        0.25 * (1 - xi) / 0.2
                    )
                )
            )
            for xi in x
        ]
    ).subdivide_sections(10)


def test_derivatives_at_zero_geometric_angles(
        AeroAnalysis: Type = asb.AeroBuildup,
):

    fuselage = make_fuselage(
        alpha_geometric=0.,
        beta_geometric=0.
    )

    airplane = asb.Airplane(
        fuselages=[fuselage],
        s_ref=np.pi,
        c_ref=1,
        b_ref=0.5,
    )

    analysis = AeroAnalysis(
        airplane=airplane,
        op_point=asb.OperatingPoint(
            velocity=100,
            alpha=0,
            beta=0,
        ),
        xyz_ref=np.array([0, 0, 0]),
    )

    try:
        aero = analysis.run_with_stability_derivatives()
    except AttributeError:
        aero = analysis.run()

    print(f"Aerodynamic coefficients with {AeroAnalysis.__name__}:")
    for key in ["Cm", "Cma", "Cn", "Cnb"]:
        print(f"{key.rjust(10)}: {aero[key]:20.4f}")

    assert aero["Cm"] == pytest.approx(0, abs=1e-3)
    assert aero["Cn"] == pytest.approx(0, abs=1e-3)

    assert aero["Cma"] == pytest.approx(
        2 * fuselage.volume() / airplane.s_ref / airplane.c_ref,
        rel=rtol
    )
    assert aero["Cnb"] == pytest.approx(
        -2 * fuselage.volume() / airplane.s_ref / airplane.b_ref,
        rel=rtol
    )


def test_derivatives_at_constant_offset(
        AeroAnalysis: Type = asb.AeroBuildup,
):

    fuselage = make_fuselage(
        alpha_geometric=5.,
        beta_geometric=10.
    )

    airplane = asb.Airplane(
        fuselages=[fuselage],
        s_ref=np.pi,
        c_ref=1.,
        b_ref=0.5,
    )

    analysis = AeroAnalysis(
        airplane=airplane,
        op_point=asb.OperatingPoint(
            velocity=100.,
            alpha=-5.,
            beta=-10.,
        ),
        xyz_ref=np.array([0, 0, 0]),
    )

    try:
        aero = analysis.run_with_stability_derivatives()
    except AttributeError:
        aero = analysis.run()

    print(f"Aerodynamic coefficients with {AeroAnalysis.__name__}:")
    for key in ["Cm", "Cma", "Cn", "Cnb"]:
        print(f"{key.rjust(10)}: {aero[key]:20.4f}")

    # assert aero["Cm"] == pytest.approx(0, abs=1e-3)
    # assert aero["Cn"] == pytest.approx(0, abs=1e-3)

    assert aero["Cma"] == pytest.approx(
        2 * fuselage.volume() / airplane.s_ref / airplane.c_ref,
        rel=rtol
    )
    assert aero["Cnb"] == pytest.approx(
        -2 * fuselage.volume() / airplane.s_ref / airplane.b_ref,
        rel=rtol
    )


def test_derivatives_at_incidence(
        AeroAnalysis: Type = asb.AeroBuildup,
):

    fuselage = make_fuselage(
        alpha_geometric=0.,
        beta_geometric=0.
    )

    airplane = asb.Airplane(
        fuselages=[fuselage],
        s_ref=np.pi,
        c_ref=1.,
        b_ref=0.5,
    )

    analysis = AeroAnalysis(
        airplane=airplane,
        op_point=asb.OperatingPoint(
            velocity=100.,
            alpha=5.,
            beta=5.,
        ),
        xyz_ref=np.array([0, 0, 0]),
    )

    try:
        aero = analysis.run_with_stability_derivatives()
    except AttributeError:
        aero = analysis.run()

    print(f"Aerodynamic coefficients with {AeroAnalysis.__name__}:")
    for key in ["Cm", "Cma", "Cn", "Cnb", "CL", "CY", "CD"]:
        print(f"{key.rjust(10)}: {aero[key]:20.4f}")

    assert aero["Cm"] == pytest.approx(
        2 * fuselage.volume() / airplane.s_ref / airplane.c_ref * np.radians(5),
        rel=rtol
    )
    assert aero["Cn"] == pytest.approx(
        -2 * fuselage.volume() / airplane.s_ref / airplane.b_ref * np.radians(5),
        rel=rtol
    )

    assert aero["Cma"] == pytest.approx(
        2 * fuselage.volume() / airplane.s_ref / airplane.c_ref,
        rel=rtol
    )
    assert aero["Cnb"] == pytest.approx(
        -2 * fuselage.volume() / airplane.s_ref / airplane.b_ref,
        rel=rtol
    )


def test_cambered_fuselage(
        AeroAnalysis: Type = asb.AeroBuildup,
):
    fuselage = asb.Fuselage(
        xsecs=[
            asb.FuselageXSec(
                xyz_c=[
                    xi,
                    0,
                    xi * (1 - xi) * 4 * 0.15
                ],
                radius=xi ** 0.5 * (1 - xi) * 0.2
            )
            for xi in np.sinspace(0, 1, 20)
        ]
    )
    # fuselage.draw_three_view()

    airplane = asb.Airplane(
        fuselages=[fuselage],
        s_ref=1,
        c_ref=1,
        b_ref=1,
    )

    analysis = AeroAnalysis(
        airplane=airplane,
        op_point=asb.OperatingPoint(
            velocity=100,
            alpha=5,
            beta=5,
        ),
        xyz_ref=np.array([0.5, 0, 0]),
    )

    try:
        aero = analysis.run_with_stability_derivatives()
    except AttributeError:
        aero = analysis.run()

    print(f"Aerodynamic coefficients with {AeroAnalysis.__name__}:")
    for key in ["Cm", "Cma", "Cn", "Cnb"]:
        print(f"{key.rjust(10)}: {aero[key]:20.4f}")

    # assert aero["Cm"] == pytest.approx(0, abs=1e-3)
    # assert aero["Cn"] == pytest.approx(0, abs=1e-3)

    assert aero["Cma"] == pytest.approx(
        2 * fuselage.volume() / airplane.s_ref / airplane.c_ref,
        rel=0.5  # this is not-strictly-speaking exact, hence the large tolerance
    )
    assert aero["Cnb"] == pytest.approx(
        -2 * fuselage.volume() / airplane.s_ref / airplane.b_ref,
        rel=0.5  # this is not-strictly-speaking exact, hence the large tolerance
    )


def test_fuselage_with_base_drag(
        AeroAnalysis: Type = asb.AeroBuildup,
):
    def r(x): # radius as a function of distance along the fuselage, x
        return 0.1 * x ** 0.5 * (2 - x)

    def z(x): # vertical displacement (in geometry axes) of the centerline, as a function of distance along fuselage, x
        return 0.15 * x * (2 - x)

    def dzdx(x): # derivative of z(x) with respect to x
        return 0.15 * ((2 - x) - x)

    def A(x): # cross-sectional area as a function of distance along fuselage, x
        return np.pi * r(x) ** 2

    fuselage = asb.Fuselage(
        xsecs=[
            asb.FuselageXSec(
                xyz_c=[
                    xi,
                    0,
                    z(xi)
                ],
                radius=r(xi)
            )
            for xi in np.sinspace(0, 1, 100)
        ]
    )
    fuselage.draw_three_view()

    airplane = asb.Airplane(
        fuselages=[fuselage],
        s_ref=1,
        c_ref=1,
        b_ref=1,
    )

    analysis = AeroAnalysis(
        airplane=airplane,
        op_point=asb.OperatingPoint(
            velocity=100,
            alpha=0,
            beta=0,
        ),
        xyz_ref=np.array([0.0, 0, 0]),
    )

    try:
        aero = analysis.run_with_stability_derivatives()
    except AttributeError:
        aero = analysis.run()

    print(f"Aerodynamic coefficients with {AeroAnalysis.__name__}:")
    for key in ["CL", "Cm", "Cma", "Cn", "Cnb"]:
        print(f"{key.rjust(10)}: {aero[key]:20.4f}")

    # TODO add assertions

    ##### Slender body theory
    l = 1 # length of fuselage
    alpha_rad = np.radians(analysis.op_point.alpha)

    from scipy import integrate

    def alpha_tilde(x):
        return alpha_rad - dzdx(x)

    CL = (
        2 / airplane.s_ref * A(l) * alpha_tilde(l)
    )
    print(f"CL_slb: {CL}")
    assert aero["CL"] == pytest.approx(CL, abs=0.01)

    Cm = (
        2 / airplane.s_ref / airplane.c_ref *
        (
            (fuselage.volume() - l * A(l)) * alpha_rad + l * A(l) * dzdx(l)
            - integrate.quad(lambda xi: A(xi) * dzdx(xi), 0, l)[0]
        )
    )

    print(f"Cm_slb: {Cm}")
    assert aero["Cm"] == pytest.approx(Cm, rel=0.5) # this is not-strictly-speaking exact, hence the large tolerance
    assert aero["Cn"] == pytest.approx(0, abs=1e-3)

    Cma = (
        2 / airplane.s_ref / airplane.c_ref *
        (fuselage.volume() - l * A(l))
    )

    # print(f"Cma_slb: {Cma}")
    # assert aero["Cma"] == pytest.approx(
    #     Cma,
    #     rel=0.5
    # )
    # assert aero["Cnb"] == pytest.approx(
    #     -2 * fuselage.volume() / airplane.s_ref / airplane.b_ref,
    #     rel=0.5 # this is not-strictly-speaking exact, hence the large tolerance
    # )


if __name__ == '__main__':
    pass
    # asb.Airplane(fuselages=[make_fuselage()]).draw_three_view()
    # # test_derivatives_at_zero_geometric_angles(asb.AVL)
    # test_derivatives_at_zero_geometric_angles(asb.AeroBuildup)
    # # test_derivatives_at_constant_offset(asb.AVL)
    # test_derivatives_at_constant_offset(asb.AeroBuildup)
    # # test_derivatives_at_incidence(asb.AVL)
    # test_derivatives_at_incidence(asb.AeroBuildup)
    # # test_cambered_fuselage(asb.AVL)
    # test_cambered_fuselage(asb.AeroBuildup)
    # test_fuselage_with_base_drag(asb.AVL)
    test_fuselage_with_base_drag(asb.AeroBuildup)



import matplotlib.pyplot as plt
from matplotlib import ticker as mt
from typing import Union, List
from aerosandbox.tools.pretty_plots.labellines import labelLines
import aerosandbox.numpy as np
from aerosandbox.tools.pretty_plots.threedim import ax_is_3d
from functools import partial
from pathlib import Path
from aerosandbox.tools import string_formatting as sf


def show_plot(
        title: str = None,
        xlabel: str = None,
        ylabel: str = None,
        zlabel: str = None,
        dpi: float = None,
        savefig: Union[str, List[str]] = None,
        savefig_transparent: bool = False,
        tight_layout: bool = True,
        tight_layout_pad: float = 0.25,
        legend: bool = None,
        legend_inline: bool = False,
        legend_frame: bool = True,
        pretty_grids: bool = True,
        set_ticks: bool = True,
        rotate_axis_labels: bool = True,
        rotate_axis_labels_linewidth: int = 14,
        show: bool = True,
):
    """
    Makes a matplotlib Figure (and all its constituent Axes) look "nice", then displays it.

    Arguments control whether various changes (from the default matplotlib settings) are made to the plot.

    One argument in particular, `show` (a boolean), controls whether the plot is displayed.

    Args:

        title: If given, sets the title of the plot. If the Figure has multiple axes, this sets the Figure-level
            suptitle instead of setting the individual Axis title.

        xlabel: If given, sets the xlabel of all axes. (Equivalent to `ax.set_xlabel(my_label)`)

        ylabel: If given, sets the ylabel of all axes. (Equivalent to `ax.set_ylabel(my_label)`)

        zlabel: If given, sets the zlabel of all axes, if the axis is 3D. (Equivalent to `ax.set_zlabel(my_label)`)

        dpi: If given, sets the dpi (display resolution, in Dots Per Inch) of the Figure.

        savefig: If given, saves the figure to the given path(s).

            * If a string is given, saves the figure to that path.
                (E.g., `savefig="my_plot.png"`)

            * If a list of strings is given, saves the figure to each of those paths.
                (E.g., `savefig=["my_plot.png", "my_plot.pdf"]`)

        savefig_transparent: If True, saves the figure with a transparent background. If False, saves the figure with
            a white background. Only has an effect if `savefig` is not None.

        tight_layout: If True, calls `plt.tight_layout()` to adjust the spacing of individual Axes. If False, skips
            this step.

        tight_layout_pad: If `tight_layout=True`, sets the padding width between the figure edge and the edges of the
        subplots, as a fraction of the font size. The default here is 0.25, which is well below the default of 1.08 in
        matplotlib. This causes Matplotlib to waste less whitespace on padding.

        legend: This value can be True, False, or None.

            * If True, displays a legend on the current Axis.

            * If False, does not add a legend on the current Axis. (However, does not delete any existing legends.)

            * If None (default), goes through some logic to determine whether a legend should be displayed. If there
                is only one line on the current Axis, no legend is displayed. If there are multiple lines, a legend is (
                in general) displayed.

        legend_inline: Boolean that controls whether an "inline" legend is displayed.

            * If True, displays an "inline" legend, where the labels are next to the lines instead of in a box.

            * If False (default), displays a traditional legend.

            Only has an effect if `legend=True` (or `legend=None`, and logic determines that a legend should be
            displayed).

        legend_frame: Boolean that controls whether a frame (rectangular background box) is displayed around the legend.
            Default is True.

        pretty_grids: Boolean that controls whether the gridlines are formatted have linewidths that are (subjectively)
            more readable.

        set_ticks: Boolean that controls whether the tick and grid locations + labels are formatted to be (
            subjectively) more readable.

            Works with both linear and log scales, and with both 2D and 3D plots.

        show: Boolean that controls whether the plot is displayed after all plot changes are applied. Default is
            True. You may want to set this to False if you want to make additional manual changes to the plot before
            displaying it. Default is True.

    Returns: None (completely in-place function). If `show=True` (default), displays the plot after applying changes.

    """
    fig = plt.gcf()
    axes = fig.get_axes()
    axes_with_3D = [ax for ax in axes if ax_is_3d(ax)]

    if pretty_grids:
        for ax in axes:
            if not ax.get_label() == '<colorbar>':
                if not ax_is_3d(ax):
                    if any(line.get_visible() for line in ax.get_xgridlines()):
                        ax.grid(True, 'major', axis='x', linewidth=1.6)
                        ax.grid(True, 'minor', axis='x', linewidth=0.7)
                    if any(line.get_visible() for line in ax.get_ygridlines()):
                        ax.grid(True, 'major', axis='y', linewidth=1.6)
                        ax.grid(True, 'minor', axis='y', linewidth=0.7)
                else:
                    for i_ax in [ax.xaxis, ax.yaxis, ax.zaxis]:
                        i_ax._axinfo["grid"].update(dict(
                            linewidth=0.7,
                        ))
                        i_ax.set_tick_params(which="minor", color=(0, 0, 0, 0))

    if set_ticks:
        for ax in axes:

            individual_axes_and_limits = {
                ax.xaxis: ax.get_xlim(),
                ax.yaxis: ax.get_ylim(),
            }
            if hasattr(ax, "zaxis"):
                individual_axes_and_limits[ax.zaxis] = ax.get_zlim()

            for i_ax, lims in individual_axes_and_limits.items():

                maj_loc = None
                maj_fmt = None
                min_loc = None
                min_fmt = None

                if i_ax.get_scale() == "log":

                    def linlogfmt(x, pos, ticks=None, default="", base=10):
                        if ticks is None:
                            ticks = [1.]

                        if x < 0:
                            sign_string = "-"
                            x = -x
                        else:
                            sign_string = ""

                        exponent = np.floor(np.log(x) / np.log(base))
                        coeff = x / base ** exponent

                        ### Fix any floating-point error during the floor function
                        if coeff < 1:
                            coeff *= base
                            exponent -= 1
                        elif coeff >= base:
                            coeff /= base
                            exponent += 1

                        for tick in ticks:
                            if np.isclose(coeff, tick):
                                return r"$\mathdefault{%s%g}$" % (
                                    sign_string,
                                    x
                                )

                        return default

                    def logfmt(x, pos, ticks=[1.], default="", base=10):
                        if x < 0:
                            sign_string = "-"
                            x = -x
                        else:
                            sign_string = ""

                        exponent = np.floor(np.log(x) / np.log(base))
                        coeff = x / base ** exponent

                        ### Fix any floating-point error during the floor function
                        if coeff < 1:
                            coeff *= base
                            exponent -= 1
                        elif coeff >= base:
                            coeff /= base
                            exponent += 1

                        for tick in ticks:
                            if tick == 1:
                                if np.isclose(coeff, 1):
                                    return r"$\mathdefault{%s%s^{%d}}$" % (
                                        sign_string,
                                        base,
                                        exponent
                                    )
                            else:
                                if np.isclose(coeff, tick):
                                    # return f"${base:.0f} {{\\times 10^{int(exponent)}}}$"
                                    return r"$\mathdefault{%s%g\times%s^{%d}}$" % (
                                        sign_string,
                                        coeff,
                                        base,
                                        exponent
                                    )
                        return default

                    ratio = lims[1] / lims[0]

                    i_ax.set_tick_params(which="minor", labelsize=8)

                    if ratio < 10:
                        maj_loc = mt.MaxNLocator(
                            nbins=6,
                            steps=[1, 2, 5, 10],
                            min_n_ticks=4,
                        )
                        maj_fmt = mt.ScalarFormatter()

                        class LogAutoMinorLocator(mt.AutoMinorLocator):
                            """
                            Dynamically find minor tick positions based on the positions of
                            major ticks. The scale must be linear with major ticks evenly spaced.
                            """

                            def __call__(self):
                                majorlocs = self.axis.get_majorticklocs()
                                try:
                                    majorstep = majorlocs[1] - majorlocs[0]
                                except IndexError:
                                    # Need at least two major ticks to find minor tick locations
                                    # TODO: Figure out a way to still be able to display minor
                                    # ticks without two major ticks visible. For now, just display
                                    # no ticks at all.
                                    return []

                                if self.ndivs is None:

                                    majorstep_no_exponent = 10 ** (np.log10(majorstep) % 1)

                                    if np.isclose(majorstep_no_exponent, [1.0, 2.5, 5.0, 10.0]).any():
                                        ndivs = 5
                                    else:
                                        ndivs = 4
                                else:
                                    ndivs = self.ndivs

                                minorstep = majorstep / ndivs

                                vmin, vmax = self.axis.get_view_interval()
                                if vmin > vmax:
                                    vmin, vmax = vmax, vmin

                                t0 = majorlocs[0]
                                tmin = ((vmin - t0) // minorstep + 1) * minorstep
                                tmax = ((vmax - t0) // minorstep + 1) * minorstep
                                locs = np.arange(tmin, tmax, minorstep) + t0

                                return self.raise_if_exceeds(locs)

                        min_loc = LogAutoMinorLocator()
                        min_fmt = mt.NullFormatter()

                    elif ratio < 10 ** 1.5:
                        maj_loc = mt.LogLocator(subs=np.arange(1, 10))
                        # if i_ax.axis_name == "x":
                        #     default = r"$^{^|}$"
                        # elif i_ax.axis_name == "y":
                        #     default = r"–"
                        # else:
                        #     default = ""

                        maj_fmt = mt.FuncFormatter(
                            partial(linlogfmt, ticks=[1, 2, 5])
                        )
                        min_loc = mt.LogLocator(numticks=999, subs=np.arange(10, 100) / 10)
                        min_fmt = mt.NullFormatter()
                    elif ratio < 10 ** 2.5:
                        maj_loc = mt.LogLocator()
                        maj_fmt = mt.FuncFormatter(partial(logfmt, ticks=[1]))
                        min_loc = mt.LogLocator(numticks=999, subs=np.arange(1, 10))
                        min_fmt = mt.FuncFormatter(partial(logfmt, ticks=[2, 5]))
                    elif ratio < 10 ** 8:
                        maj_loc = mt.LogLocator()
                        maj_fmt = mt.FuncFormatter(partial(logfmt, ticks=[1]))
                        min_loc = mt.LogLocator(numticks=999, subs=np.arange(1, 10))
                        min_fmt = mt.FuncFormatter(partial(logfmt, ticks=[1]))
                    elif ratio < 10 ** 16:
                        maj_loc = mt.LogLocator()
                        maj_fmt = mt.LogFormatterSciNotation()
                        min_loc = mt.LogLocator(numticks=999, subs=np.arange(1, 10))
                        min_fmt = mt.NullFormatter()
                    else:
                        pass

                elif i_ax.get_scale() == "linear":
                    maj_loc = mt.MaxNLocator(
                        nbins='auto',
                        steps=[1, 2, 5, 10],
                        min_n_ticks=3,
                    )

                    min_loc = mt.AutoMinorLocator()

                else:  # For any other scale, just use the default tick locations
                    continue

                if len(i_ax.get_major_ticks()) != 0:  # Unless the user has manually set the ticks to be empty
                    if maj_loc is not None:
                        i_ax.set_major_locator(maj_loc)
                    if min_loc is not None:
                        i_ax.set_minor_locator(min_loc)

                if maj_fmt is not None:
                    i_ax.set_major_formatter(maj_fmt)
                if min_fmt is not None:
                    i_ax.set_minor_formatter(min_fmt)

    ### Determine if a legend should be shown
    if legend is None:
        legend = False

        for ax in axes:
            lines = ax.get_lines()
            if len(lines) > 1:
                if not all(line.get_label()[0] == "_" for line in lines):
                    legend = True
                    break

    # Make axis labels if needed
    if xlabel is not None:
        for ax in axes:
            if not ax.get_label() == '<colorbar>':
                ax.set_xlabel(xlabel)
    if ylabel is not None:
        for ax in axes:
            if not ax.get_label() == '<colorbar>':
                ax.set_ylabel(ylabel)
    if zlabel is not None:
        if len(axes_with_3D) == 0:
            import warnings
            warnings.warn(
                "You specified a `zlabel`, but there are no 3D axes in this figure. Ignoring `zlabel`.",
                stacklevel=2
            )

        for ax in axes_with_3D:
            if not ax.get_label() == '<colorbar>':
                ax.set_zlabel(zlabel)

    # Rotate axis labels if needed
    if rotate_axis_labels:
        for ax in axes:
            if not ax_is_3d(ax):
                if not ax.get_label() == '<colorbar>':

                    ylabel = ax.get_ylabel()

                    if (rotate_axis_labels_linewidth is not None) and ("\n" not in ylabel):
                        ylabel = sf.wrap_text_ignoring_mathtext(
                            ylabel,
                            width=rotate_axis_labels_linewidth,
                        )

                    ax.set_ylabel(
                        ylabel,
                        rotation=0,
                        ha="right",
                        va="center",
                    )

    if title is not None:
        if len(axes) > 1:
            plt.suptitle(title)
        else:
            plt.title(title)

    if tight_layout:
        plt.tight_layout(
            pad=tight_layout_pad,
        )

    if legend:
        if legend_inline:  # Display an inline (matplotlib-label-lines) legend instead
            for ax in axes:
                labelLines(
                    lines=ax.get_lines(),
                )
        else:  # Display a traditional legend on the last axis
            plt.legend(frameon=legend_frame)
    if dpi is not None:
        fig.set_dpi(dpi)
    if savefig is not None:

        if not isinstance(savefig, (list, tuple, set)):
            savefig: List[Union[str, Path]] = [savefig]

        for savefig_i in savefig:
            plt.savefig(savefig_i, dpi=dpi, transparent=savefig_transparent)

    if show:
        plt.show()


def set_ticks(
        x_major: Union[float, int] = None,
        x_minor: Union[float, int] = None,
        y_major: Union[float, int] = None,
        y_minor: Union[float, int] = None,
        z_major: Union[float, int] = None,
        z_minor: Union[float, int] = None,
):
    ax = plt.gca()
    if x_major is not None:
        ax.xaxis.set_major_locator(mt.MultipleLocator(base=x_major))
    if x_minor is not None:
        ax.xaxis.set_minor_locator(mt.MultipleLocator(base=x_minor))
    if y_major is not None:
        ax.yaxis.set_major_locator(mt.MultipleLocator(base=y_major))
    if y_minor is not None:
        ax.yaxis.set_minor_locator(mt.MultipleLocator(base=y_minor))
    if z_major is not None:
        ax.zaxis.set_major_locator(mt.MultipleLocator(base=z_major))
    if z_minor is not None:
        ax.zaxis.set_minor_locator(mt.MultipleLocator(base=z_minor))


def equal() -> None:
    """
    Sets all axes to be equal. Works for both 2d plots and 3d plots.

    Returns: None

    """
    ax = plt.gca()

    if not ax_is_3d(ax):
        ax.set_aspect("equal", adjustable='box')

    else:
        ax.set_box_aspect((1, 1, 1))

        xlim = ax.get_xlim3d()
        ylim = ax.get_ylim3d()
        zlim = ax.get_zlim3d()

        x_range = abs(xlim[1] - xlim[0])
        x_middle = np.mean(xlim)
        y_range = abs(ylim[1] - ylim[0])
        y_middle = np.mean(ylim)
        z_range = abs(zlim[1] - zlim[0])
        z_middle = np.mean(zlim)

        # The plot bounding box is a sphere in the sense of the infinity
        # norm, hence I call half the max range the plot radius.
        plot_radius = 0.5 * max([x_range, y_range, z_range])

        ax.set_xlim3d([x_middle - plot_radius, x_middle + plot_radius])
        ax.set_ylim3d([y_middle - plot_radius, y_middle + plot_radius])
        ax.set_zlim3d([z_middle - plot_radius, z_middle + plot_radius])



import aerosandbox.numpy as np
from matplotlib import path
from aerosandbox.common import AeroSandboxObject
from typing import Union


class Polygon(AeroSandboxObject):
    def __init__(self,
                 coordinates: np.ndarray
                 ):
        """
        Creates a polygon object.

        Args:
            coordinates: An Nx2 NumPy ndarray of [x, y] coordinates for the polygon.
        """
        self.coordinates = np.array(coordinates)

    def __repr__(self):
        return f"Polygon ({self.n_points()} points)"

    def x(self) -> np.ndarray:
        """
        Returns the x coordinates of the polygon. Equivalent to Polygon.coordinates[:,0].

        Returns:
            X coordinates as a vector
        """
        return self.coordinates[:, 0]

    def y(self) -> np.ndarray:
        """
        Returns the y coordinates of the polygon. Equivalent to Polygon.coordinates[:,1].

        Returns:
            Y coordinates as a vector
        """
        return self.coordinates[:, 1]

    def n_points(self) -> int:
        """
        Returns the number of points/vertices/coordinates of the polygon.
        """
        try:
            return len(self.coordinates)
        except TypeError:
            try:
                return self.coordinates.shape[0]
            except AttributeError:
                return 0

    def scale(self,
              scale_x: float = 1.,
              scale_y: float = 1.,
              ) -> 'Polygon':
        """
        Scales a Polygon about the origin.
        Args:
            scale_x: Amount to scale in the x-direction.
            scale_y: Amount to scale in the y-direction.

        Returns: The scaled Polygon.
        """
        x = self.x() * scale_x
        y = self.y() * scale_y

        return Polygon(
            coordinates=np.stack((x, y), axis=1)
        )

    def translate(self,
                  translate_x: float = 0.,
                  translate_y: float = 0.,
                  ) -> 'Polygon':
        """
        Translates a Polygon by a given amount.
        Args:
            translate_x: Amount to translate in the x-direction
            translate_y: Amount to translate in the y-direction

        Returns: The translated Polygon.

        """
        x = self.x() + translate_x
        y = self.y() + translate_y

        return Polygon(
            coordinates=np.stack((x, y), axis=1)
        )

    def rotate(self,
               angle: float,
               x_center: float = 0.,
               y_center: float = 0.
               ) -> 'Polygon':
        """
        Rotates a Polygon clockwise by the specified amount, in radians.

        Rotates about the point (x_center, y_center), which is (0, 0) by default.

        Args:
            angle: Angle to rotate, counterclockwise, in radians.

            x_center: The x-coordinate of the center of rotation.

            y_center: The y-coordinate of the center of rotation.

        Returns: The rotated Polygon.

        """
        ### Translate
        translation = np.reshape(np.array([x_center, y_center]), (1, 2))
        coordinates = self.coordinates - translation

        ### Rotate
        rotation_matrix = np.rotation_matrix_2D(
            angle=angle,
        )
        coordinates = (rotation_matrix @ coordinates.T).T

        ### Translate
        coordinates = coordinates + translation

        return Polygon(
            coordinates=coordinates
        )

    def area(self) -> float:
        """
        Returns the area of the polygon.
        """
        x = self.x()
        y = self.y()
        x_n = np.roll(x, -1)  # x_next, or x_i+1
        y_n = np.roll(y, -1)  # y_next, or y_i+1

        a = x * y_n - x_n * y  # a is the area of the triangle bounded by a given point, the next point, and the origin.

        A = 0.5 * np.sum(a)  # area

        return A

    def perimeter(self) -> float:
        """
        Returns the perimeter of the polygon.
        """
        dx = np.diff(self.x())
        dy = np.diff(self.y())
        ds = (
                     dx ** 2 +
                     dy ** 2
             ) ** 0.5

        return np.sum(ds)

    def centroid(self) -> np.ndarray:
        """
        Returns the centroid of the polygon as a 1D np.ndarray of length 2.
        """
        x = self.x()
        y = self.y()
        x_n = np.roll(x, -1)  # x_next, or x_i+1
        y_n = np.roll(y, -1)  # y_next, or y_i+1

        a = x * y_n - x_n * y  # a is the area of the triangle bounded by a given point, the next point, and the origin.

        A = 0.5 * np.sum(a)  # area

        x_c = 1 / (6 * A) * np.sum(a * (x + x_n))
        y_c = 1 / (6 * A) * np.sum(a * (y + y_n))
        centroid = np.array([x_c, y_c])

        return centroid

    def Ixx(self):
        """
        Returns the nondimensionalized Ixx moment of inertia, taken about the centroid.
        """
        x = self.x()
        y = self.y()
        x_n = np.roll(x, -1)  # x_next, or x_i+1
        y_n = np.roll(y, -1)  # y_next, or y_i+1

        a = x * y_n - x_n * y  # a is the area of the triangle bounded by a given point, the next point, and the origin.

        A = 0.5 * np.sum(a)  # area

        x_c = 1 / (6 * A) * np.sum(a * (x + x_n))
        y_c = 1 / (6 * A) * np.sum(a * (y + y_n))
        centroid = np.array([x_c, y_c])

        Ixx = 1 / 12 * np.sum(a * (y ** 2 + y * y_n + y_n ** 2))

        Iuu = Ixx - A * centroid[1] ** 2

        return Iuu

    def Iyy(self):
        """
        Returns the nondimensionalized Iyy moment of inertia, taken about the centroid.
        """
        x = self.x()
        y = self.y()
        x_n = np.roll(x, -1)  # x_next, or x_i+1
        y_n = np.roll(y, -1)  # y_next, or y_i+1

        a = x * y_n - x_n * y  # a is the area of the triangle bounded by a given point, the next point, and the origin.

        A = 0.5 * np.sum(a)  # area

        x_c = 1 / (6 * A) * np.sum(a * (x + x_n))
        y_c = 1 / (6 * A) * np.sum(a * (y + y_n))
        centroid = np.array([x_c, y_c])

        Iyy = 1 / 12 * np.sum(a * (x ** 2 + x * x_n + x_n ** 2))

        Ivv = Iyy - A * centroid[0] ** 2

        return Ivv

    def Ixy(self):
        """
        Returns the nondimensionalized product of inertia, taken about the centroid.
        """
        x = self.x()
        y = self.y()
        x_n = np.roll(x, -1)  # x_next, or x_i+1
        y_n = np.roll(y, -1)  # y_next, or y_i+1

        a = x * y_n - x_n * y  # a is the area of the triangle bounded by a given point, the next point, and the origin.

        A = 0.5 * np.sum(a)  # area

        x_c = 1 / (6 * A) * np.sum(a * (x + x_n))
        y_c = 1 / (6 * A) * np.sum(a * (y + y_n))
        centroid = np.array([x_c, y_c])

        Ixy = 1 / 24 * np.sum(a * (x * y_n + 2 * x * y + 2 * x_n * y_n + x_n * y))

        Iuv = Ixy - A * centroid[0] * centroid[1]

        return Iuv

    def J(self):
        """
        Returns the nondimensionalized polar moment of inertia, taken about the centroid.
        """
        x = self.x()
        y = self.y()
        x_n = np.roll(x, -1)  # x_next, or x_i+1
        y_n = np.roll(y, -1)  # y_next, or y_i+1

        a = x * y_n - x_n * y  # a is the area of the triangle bounded by a given point, the next point, and the origin.

        A = 0.5 * np.sum(a)  # area

        x_c = 1 / (6 * A) * np.sum(a * (x + x_n))
        y_c = 1 / (6 * A) * np.sum(a * (y + y_n))
        centroid = np.array([x_c, y_c])

        Ixx = 1 / 12 * np.sum(a * (y ** 2 + y * y_n + y_n ** 2))

        Iyy = 1 / 12 * np.sum(a * (x ** 2 + x * x_n + x_n ** 2))

        J = Ixx + Iyy

        return J

    def write_sldcrv(self,
                     filepath: str = None
                     ):
        """
        Writes a .sldcrv (SolidWorks curve) file corresponding to this Polygon to a filepath.

        Args:
            filepath: A filepath (including the filename and .sldcrv extension) [string]
                if None, this function returns the .sldcrv file as a string.

        Returns: None

        """
        string = "\n".join(
            [
                "%f %f 0" % tuple(coordinate)
                for coordinate in self.coordinates
            ]
        )

        if filepath is not None:
            with open(filepath, "w+") as f:
                f.write(string)

        return string

    def contains_points(self,
                        x: Union[float, np.ndarray],
                        y: Union[float, np.ndarray],
                        ) -> Union[float, np.ndarray]:
        """
        Returns a boolean array of whether some (x, y) point(s) are contained within the Polygon.

        Note: This function is unfortunately not automatic-differentiable.

        Args:
            x: x-coordinate(s) of the query points.
            y: y-coordinate(s) of the query points.

        Returns:

            A boolean array of the same size as x and y, with values corresponding to whether the points are
            inside the Polygon.

        """
        x = np.array(x)
        y = np.array(y)
        try:
            input_shape = (x + y).shape
        except ValueError as e:  # If arrays are not broadcastable
            raise ValueError("Inputs x and y could not be broadcast together!") from e

        x = np.reshape(x, (-1, 1))
        y = np.reshape(y, (-1, 1))

        points = np.hstack((x, y))

        contained = path.Path(
            vertices=self.coordinates
        ).contains_points(
            points
        )
        contained = np.array(contained).reshape(input_shape)

        return contained

    def as_shapely_polygon(self):
        """
        Returns a Shapely Polygon object representing this polygon.

        Shapely is a Python library for 2D geometry operations. While it is more powerful than this class (e.g.,
            allows for union/intersection calculation between Polygons), it is not automatic-differentiable.
        """
        import shapely
        return shapely.Polygon(self.coordinates)

    def jaccard_similarity(self,
                           other: "Polygon"
                           ):
        """
        Calculates the Jaccard similarity between this polygon and another polygon.

        Note: This function is unfortunately not automatic-differentiable.

        Args:
            other: The other polygon to compare to.

        Returns:
            The Jaccard similarity between this polygon and the other polygon.
                * 0 if the polygons are completely disjoint
                * 1 if the polygons are identical

        """
        p1 = self.as_shapely_polygon()
        p2 = other.as_shapely_polygon()
        intersection = p1.intersection(p2).area
        union = p1.area + p2.area - intersection
        similarity = intersection / union if union != 0 else 0
        return similarity

    def draw(self,
             set_equal=True,
             color=None,
             **kwargs
             ):
        """
        Draws the Polygon on the current matplotlib axis.

        Args:

            set_equal: Whether to set the aspect ratio of the plot to be equal.

            **kwargs: Keyword arguments to pass to the matplotlib.pyplot.fill function.
                See: https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.fill.html

        Returns: None (draws on the current matplotlib axis)

        """
        import matplotlib.pyplot as plt

        if color is None:
            color = plt.gca()._get_lines.get_next_color()

        plt.fill(
            self.x(),
            self.y(),
            color=color,
            alpha=0.5,
            **kwargs
        )

        if set_equal:
            plt.gca().set_aspect("equal", adjustable='box')


if __name__ == '__main__':

    theta = np.linspace(0, 2 * np.pi, 1000)
    r = np.sin(theta) * np.sqrt(np.abs(np.cos(theta))) / (np.sin(theta) + 7 / 5) - 2 * np.sin(theta) + 2
    heart = Polygon(np.stack((r * np.cos(theta), r * np.sin(theta)), axis=1))

    import matplotlib.pyplot as plt

    fig, ax = plt.subplots()
    heart.draw()
    heart.scale(0.7, 0.7).translate(2, 1).rotate(np.radians(15)).draw()
    plt.show()



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest

def test_single_point_optimization():

    opti = asb.Opti()

    af = asb.Airfoil("naca0012")

    alpha = opti.variable(init_guess=5, n_vars=1, lower_bound=-30, upper_bound=30)

    aero = af.get_aero_from_neuralfoil(
        alpha=alpha,
        Re=1e6,
        mach=0,
    )

    opti.subject_to(
        aero["CL"] == 0.5
    )

    sol = opti.solve()

    assert sol(alpha) == pytest.approx(4.52, abs=0.5)
    assert sol(aero["CL"]) == pytest.approx(0.5, abs=0.01)

def test_multi_point_optimization():

    opti = asb.Opti()

    af = asb.Airfoil("naca0012")

    alpha = opti.variable(init_guess=5, n_vars=10, lower_bound=-30, upper_bound=30)

    aero = af.get_aero_from_neuralfoil(
        alpha=alpha,
        Re=1e6,
        mach=0,
    )

    opti.subject_to(
        aero["CL"] == 0.5
    )

    sol = opti.solve()


    assert sol(alpha)[0] == pytest.approx(4.52, abs=0.5)
    assert sol(aero["CL"])[0] == pytest.approx(0.5, abs=0.01)

if __name__ == '__main__':
    test_single_point_optimization()
    test_multi_point_optimization()
    pytest.main()


import numpy as np
from scipy import interpolate


class NaturalUnivariateSpline(interpolate.PPoly):
    """
    A Natural UnivariateSpline.

    Identical to a UnivariateSpline, except that extrapolation outside the data range is constrained to be linear.

    Based on: https://bobby.gramacy.com/surrogates/splines.html

    Which paraphrases [Hastie, Tibshirani & Friedman (2017)](https://hastie.su.domains/ElemStatLearn/), Chapters 5, 7, & 8.

    """

    def __init__(self,
                 x: np.ndarray,
                 y: np.ndarray,
                 w: np.ndarray = None,
                 k: int = 3,
                 s: float = None,
                 ext=None,
                 bbox=None,
                 check_finite=None
                 ):
        """


        Args:

            x: 1-D array of independent input data. Must be increasing; must be strictly increasing if s is 0.

            y: 1-D array of dependent input data, of the same length as x.

            w: Weights for spline fitting. Must be positive. If w is None, weights are all 1. Default is None.

            k: Degree of the smoothing spline. Must be 1 <= k <= 5. k = 3 is a cubic spline. Default is 3.

            s: Positive smoothing factor used to choose the number of knots.

        Returns:



        """
        if s is None:
            m = len(x)
            s = m - (2 * m) ** 0.5  # Identical default to UnivariateSpline's `s` argument.

        ### Deprecate and warn
        import warnings
        if ext is not None:
            warnings.warn(
                "The `ext` argument is deprecated, as a NaturalUnivariateSpline implies extrapolation.",
                DeprecationWarning
            )
        if bbox is not None:
            warnings.warn(
                "The `bbox` argument is deprecated, as a NaturalUnivariateSpline implies extrapolation.",
                DeprecationWarning
            )
        if check_finite is not None:
            warnings.warn(
                "The `check_finite` argument is deprecated.",
                DeprecationWarning
            )

        ### Compute the t, c, and k parameters for a UnivariateSpline
        tck = interpolate.splrep(
            x=x,
            y=y,
            w=w,
            k=k,
            s=s,
        )

        ### Construct the spline, without natural extrapolation
        spline = interpolate.PPoly.from_spline(
            tck=tck
        )

        ### Add spline knots for natural positive extrapolation
        spline.extend(
            c=np.array(
                [[0]] * (k - 2) + [
                    [spline(spline.x[-1], 1)],
                    [spline(spline.x[-1])]
                ]),
            x=np.array([np.inf])
        )

        ### Add spline knots for natural negative extrapolation
        spline.extend(
            c=np.array(
                [[0]] * (k - 1) + [
                    [spline(spline.x[0], 1)],
                    [spline(spline.x[0])]
                ]),
            x=np.array([spline.x[0]])
        )

        ### Construct the Natural Univariate Spline
        super().__init__(
            c=spline.c,
            x=spline.x,
            extrapolate=True,
        )


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    x = np.linspace(0, 10, 20)
    y = np.sin(x)

    nus = NaturalUnivariateSpline(
        x,
        y,
        k=3,
    )

    t, c, k = interpolate.splrep(
        x,
        y,
        s=len(x) - (2 * len(x)) ** 0.5,
    )

    us = interpolate.PPoly.from_spline((t, c, k))

    x_plot = np.linspace(-5, 15, 5000)

    fig, ax = plt.subplots()
    plt.plot(x, y, ".k", label="Data")
    plt.plot(x_plot, nus(x_plot), "--", label="Natural Univariate Spline")
    plt.plot(x_plot, us(x_plot), "--", label="Univariate Spline")
    p.set_ticks(1, 1, 1, 1)
    p.show_plot()



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest
from scipy import integrate

u_e_0 = 8
v_e_0 = 0
w_e_0 = -6
speed_0 = (u_e_0 ** 2 + w_e_0 ** 2) ** 0.5
gamma_0 = np.arctan2(-w_e_0, u_e_0)
track_0 = 0

time = np.linspace(0, 2 * np.pi, 501)


def get_trajectory(
        parameterization: type = asb.DynamicsPointMass3DCartesian,
        plot=False
):
    if parameterization is asb.DynamicsPointMass3DCartesian:
        dyn = parameterization(
            mass_props=asb.MassProperties(mass=1),
            x_e=0,
            y_e=0,
            z_e=0,
            u_e=u_e_0,
            v_e=v_e_0,
            w_e=w_e_0,
        )
    elif parameterization is asb.DynamicsPointMass3DSpeedGammaTrack:
        dyn = parameterization(
            mass_props=asb.MassProperties(mass=1),
            x_e=0,
            y_e=0,
            z_e=0,
            speed=speed_0,
            gamma=gamma_0,
            track=track_0,
        )
    else:
        raise ValueError("Bad value of `parameterization`!")

    def derivatives(t, y):
        this_dyn = dyn.get_new_instance_with_state(y)
        this_dyn.add_force(
            Fy=8,
            axes="wind"
        )

        return this_dyn.unpack_state(this_dyn.state_derivatives())

    res = integrate.solve_ivp(
        fun=derivatives,
        t_span=(time[0], time[-1]),
        t_eval=time,
        y0=dyn.unpack_state(),
        method="LSODA",
        rtol=1e-9,
        atol=1e-9,
    )

    dyn = dyn.get_new_instance_with_state(res.y)

    if plot:
        import matplotlib.pyplot as plt
        import aerosandbox.tools.pretty_plots as p

        fig, ax = plt.subplots()
        p.plot_color_by_value(dyn.x_e, dyn.altitude, c=dyn.speed, colorbar=True)
        # p.equal()
        p.show_plot("Trajectory", "$x_e$", "$z_e$")

    return dyn


def test_final_position_Cartesian():
    dyn = get_trajectory(
        parameterization=asb.DynamicsPointMass3DCartesian,
    )
    assert dyn[-1].x_e == pytest.approx(0, abs=1e-6)
    assert dyn[-1].y_e == pytest.approx(0, abs=1e-6)
    assert dyn[-1].z_e == pytest.approx(2 * np.pi * -6, abs=1e-6)
    assert dyn[-1].u_e == pytest.approx(8, abs=1e-6)
    assert dyn[-1].v_e == pytest.approx(0, abs=1e-6)
    assert dyn[-1].w_e == pytest.approx(-6, abs=1e-6)


def test_final_position_SpeedGammaTrack():
    dyn = get_trajectory(
        parameterization=asb.DynamicsPointMass3DSpeedGammaTrack,
    )
    assert dyn[-1].x_e == pytest.approx(0, abs=1e-6)
    assert dyn[-1].y_e == pytest.approx(0, abs=1e-6)
    assert dyn[-1].z_e == pytest.approx(2 * np.pi * -6, abs=1e-6)
    assert dyn[-1].u_e == pytest.approx(8, abs=1e-6)
    assert dyn[-1].v_e == pytest.approx(0, abs=1e-6)
    assert dyn[-1].w_e == pytest.approx(-6, abs=1e-6)


def test_cross_compare():
    dyn1 = get_trajectory(
        parameterization=asb.DynamicsPointMass3DCartesian,
    )
    dyn2 = get_trajectory(
        parameterization=asb.DynamicsPointMass3DSpeedGammaTrack,
    )
    assert dyn1[-1].x_e == pytest.approx(dyn2[-1].x_e, abs=1e-6, rel=1e-6)
    assert dyn1[-1].y_e == pytest.approx(dyn2[-1].y_e, abs=1e-6, rel=1e-6)
    assert dyn1[-1].z_e == pytest.approx(dyn2[-1].z_e, abs=1e-6, rel=1e-6)
    assert dyn1[-1].u_e == pytest.approx(dyn2[-1].u_e, abs=1e-6, rel=1e-6)
    assert dyn1[-1].v_e == pytest.approx(dyn2[-1].v_e, abs=1e-6, rel=1e-6)
    assert dyn1[-1].w_e == pytest.approx(dyn2[-1].w_e, abs=1e-6, rel=1e-6)


if __name__ == '__main__':
    # pytest.main()
    dyn1 = get_trajectory(
        asb.DynamicsPointMass3DCartesian,
        plot=True
    )
    dyn2 = get_trajectory(
        asb.DynamicsPointMass3DSpeedGammaTrack,
        plot=True
    )



import aerosandbox as asb
import aerosandbox.numpy as np
import aerosandbox.tools.units as u
from typing import Dict, Union


# From Torenbeek: "Synthesis of Subsonic Airplane Design", 1976, Delft University Press
# Chapter 8: "Airplane Weight and Balance"

def mass_wing_simple(
        wing: asb.Wing,
        design_mass_TOGW: float,
        ultimate_load_factor: float,
        suspended_mass: float,
        main_gear_mounted_to_wing: bool = True,
) -> float:
    """
    Computes the mass of a wing of an aircraft, according to Torenbeek's "Synthesis of Subsonic 
    Airplane Design".

    This is the simple version of the wing weight model, which is found in:
    Section 8.4: Weight Prediction Data and Methods
    8.4.1: Airframe Structure
    Eq. 8-12

    A more detailed version of the wing weight model is available in the `mass_wing()` function in this same module.

    Args:

        wing: The wing object. Should be an AeroSandbox Wing object.

        design_mass_TOGW: The design takeoff gross weight of the entire aircraft [kg].

        ultimate_load_factor: The ultimate load factor of the aircraft. 1.5x the limit load factor.

        suspended_mass: The mass of the aircraft that is suspended from the wing [kg].

        main_gear_mounted_to_wing: Whether the main gear is mounted to the wing structure.

    Returns: The total mass of the wing [kg].

    """

    k_w = np.blend(
        (design_mass_TOGW - 5670) / 2000,
        6.67e-3,
        4.90e-3
    )

    span = wing.span() / np.cosd(wing.mean_sweep_angle(x_nondim=0.5))

    wing_root_thickness = wing.xsecs[0].airfoil.max_thickness() * wing.xsecs[0].chord

    return suspended_mass * (
            k_w *
            span ** 0.75 *
            (1 + (1.905 / span) ** 0.5) *
            ultimate_load_factor ** 0.55 *
            ((span / wing_root_thickness) / (suspended_mass / wing.area())) ** 0.30 *
            (1 if main_gear_mounted_to_wing else 0.95)
    )


def mass_wing_high_lift_devices(
        wing: asb.Wing,
        max_airspeed_for_flaps: float,
        flap_deflection_angle: float = 30,
        k_f1: float = 1.0,
        k_f2: float = 1.0
) -> float:
    """
    The function mass_high_lift() is designed to estimate the weight of the high-lift devices
        on an airplane wing. It uses Torenbeek's method, which is based on multiple factors
        like wing design and flap deflection.

    Args:

        wing, an instance of AeroSandbox's Wing class,

        max_airspeed_for_flaps, the maximum airspeed at which the flaps can be deployed [m/s]

        flap_deflection_angle, the angle to which the flaps can be deflected [deg]. Default value is 30 degrees.

        k_f1, configuration factor 1, with values:
                = 1.0  for single-slotted; double-slotted, fixed hinge
                = 1.15 for double: slotted, 4-bar movement; single-slotted Fowler
                = 1.3  for double-slotted Fowler
                = 1.45 for triple-slotted Fowler

        k_f2, configuration factor 2, with values:
                = 1.0  for slotted flaps with fixed vanes
                = 1.25 for double-slotted flaps with "variable geometry", i.e., extending
                           flaps with separately moving vanes or auxiliary flaps

    Returns: Mass of the wing's high-lift system only [kg]
    """
    # S_flaps represents the total area of the control surfaces (flaps) on the wing.
    S_flaps = wing.control_surface_area()

    # Wing span
    span = wing.span()

    # Sweep at 50% chord
    sweep_half_chord = wing.mean_sweep_angle(x_nondim=0.5)
    cos_sweep_half_chord = np.cosd(sweep_half_chord)

    # span_structural is the "structural" wing span, which takes into account the wing's sweep angle.
    span_structural = span / cos_sweep_half_chord

    # Airfoil thickness over chord ratio at root
    root_t_over_c = wing.xsecs[0].airfoil.max_thickness()

    # Torenbeek Eq. C-10
    k_f = k_f1 * k_f2

    mass_trailing_edge_flaps = S_flaps * (
            2.706 * k_f *
            (S_flaps * span_structural) ** (3 / 16) *
            (
                    (max_airspeed_for_flaps / 100) ** 2 *
                    np.sind(flap_deflection_angle) *
                    np.cosd(wing.mean_sweep_angle(x_nondim=1)) /
                    root_t_over_c
            ) ** (3 / 4)
    )

    mass_leading_edge_devices = 0

    mass_high_lift_devices = mass_trailing_edge_flaps + mass_leading_edge_devices

    return mass_high_lift_devices


def mass_wing_basic_structure(
        wing: asb.Wing,
        design_mass_TOGW: float,
        ultimate_load_factor: float,
        suspended_mass: float,
        never_exceed_airspeed: float,
        main_gear_mounted_to_wing: bool = True,
        strut_y_location: float = None,
        k_e: float = 0.95,
        return_dict: bool = False,
) -> Union[float, Dict[str, float]]:
    """
    Computes the mass of the basic structure of the wing of an aircraft, according to 
        Torenbeek's "Synthesis of Subsonic Airplane Design", 1976, Appendix C: "Prediction
        of Wing Structural Weight". This is the basic wing structure without movables like spoilers,
        high-lift devices, etc.

    Likely more accurate than the Raymer wing weight models.

    Args:

        wing: The wing object.

        design_mass_TOGW: The design takeoff gross weight of the entire aircraft [kg].

        ultimate_load_factor: The ultimate load factor of the aircraft [-]. 1.5x the limit load factor.

        suspended_mass: The mass of the aircraft that is suspended from the wing [kg]. It should exclude 
            any wing attachments that are not part of the wing structure.

        never_exceed_airspeed: The never-exceed airspeed of the aircraft [m/s]. Used for flutter calculations.

        main_gear_mounted_to_wing: Whether the main gear is mounted to the wing structure. Boolean.

        strut_y_location: The spanwise-location of the strut (if any), as measured from the wing root [meters]. If None,
            it is assumed that there is no strut (i.e., the wing is a cantilever beam).

        k_e: represents weight knockdowns due to bending moment relief from engines mounted in front of elastic axis.
            see Torenbeek unlabeled equations, between C-3 and C-4.
                k_e = 1.0 if engines are not wing mounted,
                k_e = 0.95 (default) two wing mounted engines in front of the elastic axis and 
                k_e = 0.90 four wing-mounted engines in front of the elastic axis

        return_dict: Whether to return a dictionary of all the intermediate values, or just the final mass. Defaults
            to False, which returns just the final mass [kg].

    Returns: If return_dict is False (default), returns a single value: the mass of the basic wing [kg]. If return_dict is
        True, returns a dictionary of all the intermediate values.

    """

    # Wing span
    span = wing.span()

    # Sweep at 50% chord
    sweep_half_chord = wing.mean_sweep_angle(x_nondim=0.5)
    cos_sweep_half_chord = np.cosd(sweep_half_chord)

    # Structural wing span
    span_structural = span / cos_sweep_half_chord

    # Airfoil thickness over chord ratio at root
    root_t_over_c = wing.xsecs[0].airfoil.max_thickness()

    # Torenbeek Eq. C-2
    # `k_no` represents penalties due to skin joints, non-tapered skin, minimum gauge, etc.
    k_no = 1 + (1.905 / span_structural) ** 0.5

    # Torenbeek Eq. C-3
    # `k_lambda` represents penalties due to taper ratio
    k_lambda = (1 + wing.taper_ratio()) ** 0.4

    # `k_uc` represents weight knockdowns due to undercarriage.
    k_uc = 1 if main_gear_mounted_to_wing else 0.95

    # Torenbeek Eq. C-4
    # `k_st` represents weight excrescence due to structural stiffness against flutter.
    k_st = (
            1 +
            9.06e-4 * (
                    (span * np.cosd(wing.mean_sweep_angle(x_nondim=0))) ** 3 /
                    design_mass_TOGW
            ) * (
                    never_exceed_airspeed / 100 / root_t_over_c
            ) ** 2 *
            cos_sweep_half_chord
    )

    # Torenbeek Eq. C-5
    # `k_b` represents weight knockdowns due to bending moment relief from strut location.
    if strut_y_location is None:
        k_b = 1
    else:
        k_b = 1 - (strut_y_location / (wing.span() / 2)) ** 2

    ### Use all the above to compute the basic wing structural mass
    mass_wing_basic = (
            4.58e-3 *
            k_no *
            k_lambda *
            k_e *
            k_uc *
            k_st *
            (
                    k_b * ultimate_load_factor * (0.8 * suspended_mass + 0.2 * design_mass_TOGW)
            ) ** 0.55 *
            span ** 1.675 *
            root_t_over_c ** -0.45 *
            cos_sweep_half_chord ** -1.325
    )

    if return_dict:
        return locals()
    else:
        return mass_wing_basic


def mass_wing_spoilers_and_speedbrakes(
        wing: asb.Wing,
        mass_basic_wing: float
) -> float:
    """
    The function mass_spoilers_and_speedbrakes() estimates the weight of the spoilers and speedbrakes
        according to Torenbeek's "Synthesis of Subsonic Airplane Design", 1976, Appendix C: "Prediction
        of Wing Structural Weight".

    N.B. the weight is coming out unrealistic and approx. 20-30% of the weight of the wing. This needs
        a correction. It uses normally the 12.2 kg/m^2 wing area.

    Args:

        wing: an instance of AeroSandbox's Wing class.

        mass_basic_wing: The basic weight of the wing (without spoilers, speedbrakes, flaps, slats) [kg]

    Returns: The mass of the spoilers and speed brakes only [kg]

    N.B. the weight estimation using the 12.2 kg/m^2 figure comes out too high if using
        the wing as a referenced area. Reduced to 1.5% of the basic wing mass.
    """
    # mass_spoilers_and_speedbrakes = np.softmax(
    #                                            12.2 * wing.area(),
    #                                            0.015 * mass_basic_wing
    #                                            )

    mass_spoilers_and_speedbrakes = 0.015 * mass_basic_wing

    return mass_spoilers_and_speedbrakes


def mass_wing(
        wing: asb.Wing,
        design_mass_TOGW: float,
        ultimate_load_factor: float,
        suspended_mass: float,
        never_exceed_airspeed: float,
        max_airspeed_for_flaps: float,
        main_gear_mounted_to_wing: bool = True,
        flap_deflection_angle: float = 30,
        strut_y_location: float = None,
        return_dict: bool = False,
) -> Union[float, Dict[str, float]]:
    """
    Computes the mass of a wing of an aircraft, according to Torenbeek's "Synthesis of Subsonic Airplane Design",
        1976, Appendix C: "Prediction of Wing Structural Weight".

    Likely more accurate than the Raymer wing weight models.

    Args:

        wing: The wing object.

        design_mass_TOGW: The design takeoff gross weight of the entire aircraft [kg].

        ultimate_load_factor: The ultimate load factor of the aircraft. 1.5x the limit load factor.

        suspended_mass: The mass of the aircraft that is suspended from the wing [kg].

        never_exceed_airspeed: The never-exceed airspeed of the aircraft [m/s]. Used for flutter calculations.

        max_airspeed_for_flaps: The maximum airspeed at which the flaps are allowed to be deployed [m/s]. In the
        absence of other information, 1.8x stall speed is a good guess.

        main_gear_mounted_to_wing: Whether the main gear is mounted to the wing structure.

        flap_deflection_angle: The maximum deflection angle of the flaps [deg].

        strut_y_location: The y-location of the strut (if any), relative to the wing's leading edge [m]. If None,
            it is assumed that there is no strut (i.e., the wing is a cantilever beam).

        return_dict: Whether to return a dictionary of all the intermediate values, or just the final mass. Defaults
            to False, which returns just the final mass.

    Returns: If return_dict is False (default), returns a single value: the total mass of the wing [kg]. If
        return_dict is True, returns a dictionary of all the intermediate values.

    """

    # High-lift mass estimation
    mass_high_lift_devices = mass_wing_high_lift_devices(
        wing=wing,
        max_airspeed_for_flaps=max_airspeed_for_flaps,
        flap_deflection_angle=flap_deflection_angle,
    )
    # Basic wing structure mass estimation
    mass_basic_wing = mass_wing_basic_structure(
        wing=wing,
        design_mass_TOGW=design_mass_TOGW,
        ultimate_load_factor=ultimate_load_factor,
        suspended_mass=suspended_mass,
        never_exceed_airspeed=never_exceed_airspeed,
        main_gear_mounted_to_wing=main_gear_mounted_to_wing,
        strut_y_location=strut_y_location,
    )
    # spoilers and speedbrake mass estimation
    mass_spoilers_speedbrakes = mass_wing_spoilers_and_speedbrakes(
        wing=wing,
        mass_basic_wing=mass_basic_wing
    )

    mass_wing_total = (
            mass_basic_wing +
            1.2 * (mass_high_lift_devices + mass_spoilers_speedbrakes)
    )

    if return_dict:
        return locals()
    else:
        return mass_wing_total


# def mass_hstab(
#         hstab: asb.Wing,
#         design_mass_TOGW: float,
#         ultimate_load_factor: float,
#         suspended_mass: float,
#         main_gear_mounted_to_wing: bool = True,
# ) -> float:
#
#     k_wt = 0.64

def mass_fuselage_simple(
        fuselage: asb.Fuselage,
        never_exceed_airspeed: float,
        wing_to_tail_distance: float,
):
    """
    Computes the mass of the fuselage, using Torenbeek's simple version of the calculation.

    Source:
    Torenbeek: "Synthesis of Subsonic Airplane Design", 1976
    Section 8.4: Weight Prediction Data and Methods
    8.4.1: Airframe Structure
    Eq. 8-16

    Args:

        fuselage: The fuselage object. Should be an AeroSandbox Fuselage object.

        never_exceed_airspeed: The never-exceed airspeed of the aircraft, in m/s.

        wing_to_tail_distance: The distance from the quarter-chord of the wing to the quarter-chord of the tail,
        in meters.

    Returns: The mass of the fuselage, in kg.

    """
    widths = [
        xsec.width
        for xsec in fuselage.xsecs
    ]

    max_width = np.softmax(
        *widths,
        softness=np.mean(np.array(widths)) * 0.01
    )

    heights = [
        xsec.height
        for xsec in fuselage.xsecs
    ]

    max_height = np.softmax(
        *heights,
        softness=np.mean(np.array(heights)) * 0.01
    )

    return (
            0.23 *
            (
                    never_exceed_airspeed *
                    wing_to_tail_distance /
                    (max_width + max_height)
            ) ** 0.5 *
            fuselage.area_wetted() ** 1.2
    )


def mass_fuselage(
        fuselage: asb.Fuselage,
        design_mass_TOGW: float,
        ultimate_load_factor: float,
        never_exceed_airspeed: float,
        wing_to_tail_distance: float,
):
    # TODO Torenbeek Appendix D (PDF page 477)

    # Stage 1: Calculate the weight of the fuselage shell, which carries the primary loads and contributes
    # approximately 1/3 to 1/2 of the fuselage weight ("gross shell weight").

    # Torenbeek Eq. D-3
    fuselage.fineness_ratio()

    fuselage_quasi_slenderness_ratio = fuselage.fineness_ratio(assumed_shape="sears_haack")

    k_lambda = np.softmin(
        0.56 * fuselage.fineness_ratio(assumed_shape="sears_haack")
    )

    W_sk = 0.05428 * k_lambda * S_g ** 1.07 * never_exceed_airspeed ** 0.743

    W_g = W_sk + W_str + W_fr


def mass_propeller(
        propeller_diameter: float,
        propeller_power: float,
        n_blades: int,
) -> float:
    """
    Computes the mass of a propeller.

    From Torenbeek: "Synthesis of Subsonic Airplane Design", 1976, Delft University Press.
    Table 8-9 (pg. 286, PDF page 306)

    Args:

        propeller_diameter: Propeller diameter, in meters.

        propeller_power: Propeller power, in watts.

        n_blades: Number of propeller blades.

    Returns: Propeller mass, in kilograms.

    """
    return (
            0.108 *
            n_blades *
            (
                    (propeller_diameter / u.foot) *
                    (propeller_power / u.horsepower)
            ) ** 0.78174
    ) * u.lbm



import sympy as s
from sympy import init_printing
init_printing()

# Reconstructs a quadratic interpolant from x1...x3, then gets the derivative at x2

# Define the symbols
x1, x2, x3 = s.symbols('x1 x2 x3', real=True)
f1, f2, f3 = s.symbols('f1 f2 f3', real=True)

# hm = x2 - x1
# hp = x3 - x2
hm, hp = s.symbols("hm hp")

q = s.symbols('q')  # Normalized space for a Bernstein basis.
# Mapping from x-space to q-space has x=x2 -> q=0, x=x3 -> q=1.

q1 = 0
# q2 = s.symbols('q2', real=True) # (x2 - x1) / (x3 - x1)
q2 = hm / (hm + hp)
q3 = 1

# Define the Bernstein basis polynomials
b1 = (1 - q) ** 2
b2 = 2 * q * (1 - q)
b3 = q ** 2

c1, c2, c3 = s.symbols('c1 c2 c3', real=True)

# Can solve for c2 and c3 exactly
c1 = f1
c3 = f3

f = c1 * b1 + c2 * b2 + c3 * b3

f2_quadratic = f.subs(q, q2)#.simplify()

factors = [q2]
# factors = [f1, f2, f3, f4]

# Solve for c2 and c3
sol = s.solve(
    [
        f2_quadratic - f2,
    ],
    [
        c2,
    ],
)
c2 = sol[c2].factor(factors).simplify()


f = c1 * b1 + c2 * b2 + c3 * b3
dfdq = f.diff(q).simplify()
# dqdx = 1 / (x3 - x1)
dqdx = 1 / (x3 - x1)
dfdx = dfdq * dqdx

dfm, dfp = s.symbols("dfm dfp")

def simplify(expr):
    import copy
    original_expr = copy.copy(expr)
    expr = expr.subs({
        f3 - f2: dfp,
        f2 - f1: dfm,
        f3 - f1: dfp + dfm,
        x3 - x1: hm + hp,
    })
    expr = expr.subs({
        f3 - f2: dfp,
        f2 - f1: dfm,
        f3 - f1: dfp + dfm,
        x3 - x1: hm + hp,
    })
    expr = expr.factor([
        hp,
        hm
    ]).simplify()
    if expr != original_expr:
        expr = simplify(expr)
    return expr

dfdx_q1 = simplify(dfdx.subs(q, q1))
dfdx_q2 = simplify(dfdx.subs(q, q2))
dfdx_q3 = simplify(dfdx.subs(q, q3))


# integral = (c1 + c2 + c3) / 3 # God I love Bernstein polynomials



# integral = s.simplify(integral)

parsimony = len(str(dfdx_q1))
print(s.pretty(dfdx_q1, num_columns=100))
print(f"Parsimony: {parsimony}")


import aerosandbox.numpy as np
from typing import Union


def linear_hermite_patch(
        x: Union[float, np.ndarray],
        x_a: float,
        x_b: float,
        f_a: float,
        f_b: float,
) -> Union[float, np.ndarray]:
    """
    Computes the linear Hermite polynomial patch that passes through the given endpoints f_a and f_b.

    Args:
        x: Scalar or array of values at which to evaluate the patch.
        x_a: The x-coordinate of the first endpoint.
        x_b: The x-coordinate of the second endpoint.
        f_a: The function value at the first endpoint.
        f_b: The function value at the second endpoint.

    Returns:
        The value of the patch evaluated at the input x. Returns a scalar if x is a scalar, or an array if x is an array.
    """
    return (x - x_a) * (f_b - f_a) / (x_b - x_a) + f_a


def cubic_hermite_patch(
        x: Union[float, np.ndarray],
        x_a: float,
        x_b: float,
        f_a: float,
        f_b: float,
        dfdx_a: float,
        dfdx_b: float,
        extrapolation: str = 'continue',
) -> Union[float, np.ndarray]:
    """
    Computes the cubic Hermite polynomial patch that passes through the given endpoints and endpoint derivatives.

    Args:
        x: Scalar or array of values at which to evaluate the patch.
        x_a: The x-coordinate of the first endpoint.
        x_b: The x-coordinate of the second endpoint.
        f_a: The function value at the first endpoint.
        f_b: The function value at the second endpoint.
        dfdx_a: The derivative of the function with respect to x at the first endpoint.
        dfdx_b: The derivative of the function with respect to x at the second endpoint.
        extrapolation: A string indicating how to handle extrapolation outside of the domain [x_a, x_b]. Valid values are
                      "continue", which continues the patch beyond the endpoints, and "clip", which clips the patch at the
                      endpoints. Default is "continue".

    Returns:
        The value of the patch evaluated at the input x. Returns a scalar if x is a scalar, or an array if x is an array.
    """
    dx = x_b - x_a
    t = (x - x_a) / dx  # Nondimensional distance along the patch
    if extrapolation == 'continue':
        pass
    elif extrapolation == 'clip':
        t = np.clip(t, 0, 1)
    else:
        raise ValueError("Bad value of `extrapolation`!")

    return (
            (t ** 3) * (1 * f_b) +
            (t ** 2 * (1 - t)) * (3 * f_b - 1 * dfdx_b * dx) +
            (t * (1 - t) ** 2) * (3 * f_a + 1 * dfdx_a * dx) +
            ((1 - t) ** 3) * (1 * f_a)
    )


def cosine_hermite_patch(
        x: Union[float, np.ndarray],
        x_a: float,
        x_b: float,
        f_a: float,
        f_b: float,
        dfdx_a: float,
        dfdx_b: float,
        extrapolation: str = 'continue',
) -> Union[float, np.ndarray]:
    """
    Computes a Hermite patch (i.e., values + derivatives at endpoints) that uses a cosine function to blend between
    linear segments.

    The end result is conceptually similar to a cubic Hermite patch, but computation is faster and the patch is
    C^\infty-continuous.

    Args:
        x: Scalar or array of values at which to evaluate the patch.
        x_a: The x-coordinate of the first endpoint.
        x_b: The x-coordinate of the second endpoint.
        f_a: The function value at the first endpoint.
        f_b: The function value at the second endpoint.
        dfdx_a: The derivative of the function with respect to x at the first endpoint.
        dfdx_b: The derivative of the function with respect to x at the second endpoint.
        extrapolation: A string indicating how to handle extrapolation outside of the domain [x_a, x_b]. Valid values are
                       "continue", which continues the patch beyond the endpoints, and "linear", which extends the patch
                       linearly at the endpoints. Default is "continue".

    Returns:
        The value of the patch evaluated at the input x. Returns a scalar if x is a scalar, or an array if x is an array.
    """
    t = (x - x_a) / (x_b - x_a)  # Nondimensional distance along the patch
    if extrapolation == 'continue':
        pass
    elif extrapolation == 'linear':
        t = np.clip(t, 0, 1)
    else:
        raise ValueError("Bad value of `extrapolation`!")

    l1 = (x - x_a) * dfdx_a + f_a
    l2 = (x - x_b) * dfdx_b + f_b

    b = 0.5 + 0.5 * np.cos(np.pi * t)

    return b * l1 + (1 - b) * l2


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    x = np.linspace(-1, 2, 500)
    plt.plot(
        x,
        cubic_hermite_patch(
            x,
            x_a=0,
            x_b=1,
            f_a=0,
            f_b=1,
            dfdx_a=-0.5,
            dfdx_b=-1,
            extrapolation='clip'
        )
    )

    p.equal()
    p.show_plot()
